'''Trace implementation for the FST file format '''

import re
import sys
import pylibfst as fst

from wal.trace.trace import Trace

class TraceFst(Trace):
    '''Holds data for one fst trace.'''

    def __init__(self, file, tid, container, from_string=False, keep_signals=None):
        super().__init__(tid, file, container)
        self.buf = fst.ffi.new("char[256]")

        if from_string:
            raise ValueError("FST traces do not support the from_string argument")

        self.fst = fst.lib.fstReaderOpen(file.encode('utf-8'))
        if self.fst == fst.ffi.NULL:
            print(f'Error while loading {file}')
            sys.exit(1)

        # get scopes and signals
        (self.scopes, signals) = fst.get_scopes_signals2(self.fst)
        self.references_to_ids = signals.by_name

        # rename grouped scopes
        self.scopes = [
            re.sub(r'\(([0-9]+)\)', r'<\1>', scope) for scope in self.scopes]
        self.scopes = [
            re.sub(r'\[([0-9]+)\]', r'<\1>', scope) for scope in self.scopes]

        # get mapping from name to tid and remove trailing signal width, ' [31:0]' etc.
        self.references_to_ids = {re.sub(r' *\[\d+:\d+\]', '', k): v for
                                        k, v in self.references_to_ids.items()}
        # rename grouped signals like reg(0), reg(1) to reg<0>, reg<1>
        self.references_to_ids = {
            re.sub(r'\(([0-9]+)\)', r'<\1>', k): v for k, v in self.references_to_ids.items()}
        # rename grouped signals like reg[0], reg[1] to reg<0>, reg<1>
        self.references_to_ids = {
            re.sub(r'\[([0-9]+)\]', r'<\1>', k): v for k, v in self.references_to_ids.items()}

        self.rawsignals = list(self.references_to_ids.keys())
        self.rawsignals_by_handle = [k for k, v in self.references_to_ids.items() if v.handle in signals.by_handle]
        self.signals = set(self.rawsignals)
        self.signals_by_handle = set(self.rawsignals_by_handle)

        # remove duplicate timestamps, enumerate all timestamps and create look up table
        fst.lib.fstReaderSetFacProcessMaskAll(self.fst)
        raw_timestamps = fst.lib.fstReaderGetTimestamps(self.fst)
        self.all_timestamps = raw_timestamps.val
        self.timestamps = self.all_timestamps

        # stores current time stamp
        self.index = 0
        self.max_index = raw_timestamps.nvals - 1

    def access_signal_data(self, name, index):
        '''Backend specific function for accessing signals in the waveform'''
        handle = self.references_to_ids[name].handle
        value = fst.helpers.string(fst.lib.fstReaderGetValueFromHandleAtTime(self.fst, self.timestamps[index], handle, self.buf))
        try:
            return int(value, 2)
        except ValueError:
            if value[0] == 'r':
                return float(value[1:])
            else:
                return value

    def set_sampling_points(self, new_indices):
        super().set_sampling_points(new_indices)

    def signal_width(self, name):
        return self.references_to_ids[name].length



'''Wrapper class for trace data'''
import sys
import pathlib

from wal.ast_defs import VirtualSignal
from wal.trace.trace import Trace
from wal.trace.vcd import TraceVcd
from wal.trace.csvtrace import TraceCsv

class TraceContainer:
    '''Can hold multiple traces and dispatches value access to the correct trace.'''

    def __init__(self):
        self.traces = {}
        self.n_traces = 0
        self.index_stack = []


    def load(self, file, tid=None, from_string=False, keep_signals=None):
        '''Load a trace from file and add it under trace id tid.'''
        if tid is None:
            tid = f't{len(self.traces)}'

        assert tid not in self.traces, f'load: trace id {tid} already in use'
        
        file_extension = pathlib.Path(file).suffix
        if file_extension == '.vcd':
            self.traces[tid] = TraceVcd(file, tid, self, from_string=from_string, keep_signals=keep_signals)
        elif file_extension == '.fst':
            try:
                from wal.trace.fst import TraceFst
                self.traces[tid] = TraceFst(file, tid, self, from_string=from_string, keep_signals=keep_signals)
            except ModuleNotFoundError:
                print(f'Can not open file "{file}"')
                print('To get support for the "fst" filetype install "pylibfst" package')
                print('More information on pylibfst: https://pypi.org/project/pylibfst/')
                sys.exit(1)
        elif file_extension == '.csv':
            self.traces[tid] = TraceCsv(file, tid, self, from_string=from_string, keep_signals=keep_signals)
        else:
            print(f'File extension "{file_extension}" not supported.')

        self.n_traces += 1


    def unload(self, tid='DEFAULT'):
        '''Remove the trace tid from the set of loaded traces. '''
        if tid in self.traces:
            del self.traces[tid]
            self.n_traces -= 1


    def signal_value(self, name, offset=0, scope=''):
        '''Get the value of signal name at current index + offset.'''
        if self.n_traces == 1 and Trace.SCOPE_SEPERATOR not in name:
            trace = list(self.traces.values())[0]
            return trace.signal_value(name, offset, scope)

        if Trace.SCOPE_SEPERATOR in name:
            # extract tid of vcd
            separator = name.index(Trace.SCOPE_SEPERATOR)
            trace_tid = name[:separator]
            signal_name = name[separator+1:]
            assert trace_tid in self.traces, f'No trace with tid {trace_tid}'
            return self.traces[trace_tid].signal_value(signal_name, offset, scope)


        raise RuntimeError('No traces loaded')

    def signal_width(self, name):
        '''Get the value of signal name at current index + offset.'''

        if self.n_traces == 1 and Trace.SCOPE_SEPERATOR not in name:
            trace = list(self.traces.values())[0]
            return trace.signal_width(name)

        if Trace.SCOPE_SEPERATOR in name:
            # extract tid of vcd
            separator = name.index(Trace.SCOPE_SEPERATOR)
            trace_tid = name[:separator]
            signal_name = name[separator+1:]
            assert trace_tid in self.traces, f'No trace with tid {trace_tid}'
            return self.traces[trace_tid].signal_width(signal_name)

        raise RuntimeError('No traces loaded')

    def contains(self, name):
        '''Return true if a signal with name exists in any trace.'''

        if self.n_traces == 1 and Trace.SCOPE_SEPERATOR not in name:
            return name in (list(self.traces.values())[0]).signals or name in Trace.SPECIAL_SIGNALS

        if Trace.SCOPE_SEPERATOR in name:
            separator = name.index(Trace.SCOPE_SEPERATOR)
            trace_tid = name[:separator]
            signal_name = name[separator+1:]
            assert trace_tid in self.traces, f'No trace with tid {trace_tid}'
            return signal_name in self.traces[trace_tid].signals or signal_name in Trace.SPECIAL_SIGNALS

        return False

    def step(self, steps=1, tid=None):
        '''Step one or all traces.
        If steps is not defined step trace(s) by +1.
        If no trace id (tid) is defined step all traces.
        '''
        ended = []
        if tid:
            assert tid in self.traces, f'No trace with tid {tid}'
            res =  self.traces[tid].step(steps)
            if res:
                ended.append(res)
        else:
            for trace in self.traces.values():
                res = trace.step(steps)
                if res:
                    ended.append(res)

        return ended


    @property
    def scopes(self):
        '''Returns a list of all scopes.
        If more than one scope trace is loaded, scopes are prepended by the tid of their trace.
        '''
        if len(self.traces) > 1:
            scopes = [(trace.id + trace.SCOPE_SEPERATOR + scope) for trace in self.traces.values() for scope in trace.scopes]
        else:
            scopes = [scope for trace in self.traces.values() for scope in trace.scopes]

        return scopes


    @property
    def signals(self):
        '''Returns a list of all signals.
        If more than one scope trace is loaded, signals are prepended by the tid of their trace
        '''
        signals = []
        for trace in self.traces.values():
            if len(self.traces) == 1:
                tmp = list(map(lambda s: f'{s}', trace.rawsignals))
            else:
                tmp = list(map(lambda s, t=trace: f'{t}{Trace.SCOPE_SEPERATOR}{s}', trace.rawsignals))
            signals = signals + tmp
        return signals


    def indices(self):
        '''returns a dict of all current indices'''
        indices = {}
        for trace in self.traces.values():
            indices[trace.tid] = trace.index

        return indices


    def store_indices(self):
        '''Store the current indices for all waveforms on a stack'''
        self.index_stack.append(self.indices())


    def restore_indices(self):
        '''Restore the indices for all waveforms from the stack'''
        if self.index_stack:
            indices = self.index_stack.pop()
            for tid, index in indices.items():
                self.traces[tid].set(index)


    def add_virtual_signal(self, name, expr, seval):
        '''Add a virtual signal to the trace'''

        if self.n_traces == 1:
            trace = list(self.traces.values())[0]
            signal = VirtualSignal(name, expr, trace, seval)
            return trace.add_virtual_signal(signal)

        if self.n_traces > 1 or Trace.SCOPE_SEPERATOR in name:
            # extract tid of vcd
            separator = name.index(Trace.SCOPE_SEPERATOR)
            trace_tid = name[:separator]
            assert trace_tid in self.traces, f'No trace with tid {trace_tid}'
            signal = VirtualSignal(name, expr, self.traces[trace_tid], seval)
            return self.traces[trace_tid].add_virtual_signal(signal)

        raise RuntimeError('No traces loaded')



'''S-Exprssion eval functions'''

import os
import wal

from importlib.resources import files

from wal.util import wal_str
from wal.ast_defs import Operator, UserOperator, Symbol, Environment, Closure, Macro, WList, WalEvalError
from wal.implementation.types import type_operators
from wal.implementation.math import math_operators
from wal.implementation.bitwise import bitwise_operators
from wal.implementation.list import list_operators
from wal.implementation.array import array_operators
from wal.implementation.wal import wal_operators
from wal.implementation.core import core_operators
from wal.implementation.special import special_operators
from wal.implementation.virtual import virtual_operators


class SEval:
    '''Evaluate s-expressions'''

    def __init__(self, traces):
        '''Initial Evaluation Object'''
        self.traces = traces
        self.global_environment = None
        self.environment = None
        self.gensymi = None
        self.imports = None
        self.aliases = None
        self.macros = None
        self.scope = None
        self.group = None
        self.context = None
        self.reset()
        self.dispatch = {**core_operators, **math_operators, **bitwise_operators, \
            **type_operators, **list_operators, **array_operators, \
            **wal_operators, **special_operators, **virtual_operators}
        self.user_dispatch = {}
        initial_walpath = ['.', os.path.expanduser('~/.wal/libs/'), str(files(wal).joinpath('libs/'))]
        self.walpath = initial_walpath + os.getenv('WALPATH', '').split(';')

    def reset(self):
        '''Resets all traces back to time 0 and resets all WAL elements (e.g. aliases, imports, ...) '''
        # roll back traces
        for trace in self.traces.traces.values():
            trace.set(0)

        self.global_environment = Environment()
        self.environment = self.global_environment
        self.gensymi = 0
        self.imports = {}
        self.aliases = {}
        self.macros = {}
        self.scope = ''
        self.group = ''
        self.virtual_signals = {}
        self.global_environment.define('CS', '')
        self.global_environment.define('CG', '')
        self.global_environment.define('ARGS', [])
        self.global_environment.define('VIRTUAL', self.virtual_signals)

    def eval_args(self, args):
        '''Helper function to evaluate each element of a list'''
        return list(map(self.eval, args))

    def eval_dispatch(self, oprtr, args):
        '''Dispatch operators'''
        return self.dispatch.get(oprtr.value, lambda a, b: NotImplementedError())(self, args)

    def eval_user_dispatch(self, oprtr, args):
        '''Dispatch operators added by the user'''
        return self.user_dispatch.get(oprtr.name, lambda a, b: NotImplementedError())(self, args)

    def eval_closure(self, closure, args):
        '''Evaluate a closure.'''
        save_env = self.environment
        new_env = Environment(parent=closure.environment)

        if isinstance(closure.args, Symbol):
            new_env.define(closure.args.name, self.eval_args(args))
        elif isinstance(closure.args, WList):
            assert len(closure.args) == len(args), f'{closure.name}: number of passed arguments does not match expected number'
            for arg, val in zip(closure.args, args):
                new_env.define(arg.name, self.eval(val))
        else:
            assert False, f'cannot evaluate {wal_str(closure)}'

        self.environment = new_env
        try:
            res = self.eval(closure.expression)
        except WalEvalError as error:
            error.add(closure.name)
            raise error

        self.environment = save_env
        return res

    def eval(self, expr):
        '''Main s-expression eval function'''
        res = NotImplementedError()
        try:
            if isinstance(expr, Symbol):
                name = expr.name
                if expr.name in self.aliases: # if an alias exists
                    name = self.aliases[expr.name]

                # this symbol was already resolved
                if expr.steps is not None:
                    env = self.environment
                    steps = expr.steps
                    while steps > 0:
                        env = self.environment.parent
                        steps -= 1

                    res = env.read(expr.name)
                elif self.traces.contains(name):  # if symbol is a signal from wavefile
                    res = self.traces.signal_value(name, scope=self.scope)
                else:
                    # this symbol has not been resolved
                    res = self.environment.read(name)
            elif isinstance(expr, (WList, list)):
                head = expr[0]
                tail = expr[1:]

                if isinstance(head, Operator):
                    res = self.eval_dispatch(head, tail)
                elif isinstance(head, Closure):
                    res = self.eval_closure(head, tail)
                elif isinstance(head, UserOperator):
                    res = self.eval_user_dispatch(head, tail)
                elif isinstance(head, Macro):
                    expanded = self.eval(head.expression)
                    expr.clear()
                    for expression in expanded:
                        expr.append(expression)
                    res = self.eval(expr)
                elif callable(expr[0]):
                    res = expr[0](*expr[1:])
                elif isinstance(head, (int, str)):
                    assert False, f'{wal_str(expr)} is not a valid function call'
                else:
                    func = self.eval(head)
                    res = self.eval([func] + tail)
            elif isinstance(expr, (int, str, float, Closure)):
                res = expr
        except AssertionError as error:
            self.print_error(expr, error)
            raise WalEvalError()

        if not isinstance(res, Exception):
            return res
        raise NotImplementedError(str(expr))


    def print_error(self, expr, error):
        print()
        print('>>>>> WAL Runtime error! <<<<<')
        if isinstance(expr, (WList, Symbol)) and expr.line_info and expr.line_info['filename'] != '':
            with open(expr.line_info['filename']) as f:
                lines = f.readlines()
                error_range = range(expr.line_info['line'] - 5, expr.line_info['end_line'])
                max_line_nr_width = max([len(str(lnr)) for lnr in error_range])

                for lnr in error_range:
                    if lnr >= 0:
                        line = lines[lnr].rstrip()
                        if lnr+1 >= expr.line_info['line'] and lnr+1 <= expr.line_info['end_line']:
                            print(f'\033[91m{lnr+1:{max_line_nr_width}}:\033[0m ', end='')
                        else:
                            print(f'{lnr+1:{max_line_nr_width}}: ', end='')

                        for cnr, char in enumerate(line):
                            if lnr + 1 == expr.line_info['line'] and (cnr + 1) == expr.line_info['column']:
                                print('\033[91m', end='')

                            print(char, end='')
                            if lnr + 1 == expr.line_info['end_line'] and cnr + 2 == expr.line_info['end_column']:
                                print('\033[0m', end='')

                        print()

                print(' ' * (max_line_nr_width + 1), end='')
                for cnr, char in enumerate(lines[expr.line_info['line'] - 1]):
                    if char.isspace():
                        print(char, end='')
                    elif cnr == expr.line_info['column']:
                        print('\033[91m^\033[0m', end='')
                    else:
                        print(' ', end='')

            print()
            if expr.line_info['filename']:
                print('Error at ', expr.line_info['filename'], ':', expr.line_info['line'], sep='')
            else:
                print('Error on line', expr.line_info['line'])

            print('\033[0m', end='')

        print(error)



'''wal command line compiler'''
import argparse
import pickle
from pathlib import Path
from wal.version import __version__

from wal.reader import read_wal_sexprs
from wal.passes import expand, optimize


class Arguments:  # pylint: disable=too-few-public-methods
    '''Wrapper class for argument parser'''

    def __init__(self):
        parser = argparse.ArgumentParser()
        self.parser = parser

        parser.add_argument('input_name', help='input filename')
        parser.add_argument('-o', action='store', dest='output_name',
                            help='output filename', default=None)
        parser.add_argument('-v', '--version', action='version',
                            version=f'%(prog)s {__version__}')

    def parse(self):
        '''Parse program arguments and check minimal requirements'''
        args = self.parser.parse_args()

        if args.input_name is None:
            return None
        return args


def wal_compile(inname, outname, wal):
    '''Compiles the file inname to outname'''
    with open(inname, 'r', encoding='utf8') as fin:
        code = fin.read()
        compiled = []

        for sexpr in read_wal_sexprs(code):
            expanded = expand(wal.eval_context, sexpr, parent=wal.eval_context.global_environment)
            optimized = optimize(expanded)
            compiled.append(optimized)

        if outname:
            name = outname
        else:
            name = Path(inname).with_suffix('.wo')

        with open(name, 'wb') as fout:
            pickle.dump(compiled, fout)


def run():  # pylint: disable=R1710
    '''Entry point for the compiler script'''

    arg_parser = Arguments()
    args = arg_parser.parse()

    from wal.core import Wal
    wal_compile(args.input_name, args.output_name, Wal())



'''Implementations for type related functions such as type checking'''
from wal.ast_defs import Operator, Symbol, WList


def op_is_defined(seval, args):
    '''Returns true if the argument is a defined symbol'''
    assert len(args) == 1, 'defined?: expects exactly one argument'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, Symbol), 'defined?: argument must evaluate to symbol'
    return bool(seval.environment.is_defined(evaluated.name))


def op_is_atom(seval, args):
    '''Returns true if all arguments evaluate to atoms'''
    evaluated = seval.eval_args(args)
    return all(isinstance(arg, (Operator, Symbol, str, int)) for arg in evaluated)


def op_is_symbol(seval, args):
    '''Returns true if all arguments evaluate to symbols'''
    evaluated = seval.eval_args(args)
    return all(isinstance(arg, Symbol) for arg in evaluated)


def op_is_string(seval, args):
    '''Returns true if all arguments evaluate to strings'''
    evaluated = seval.eval_args(args)
    return all(isinstance(arg, str) for arg in evaluated)


def op_is_int(seval, args):
    '''Returns true if all arguments evaluate to integers'''
    evaluated = seval.eval_args(args)
    return all(isinstance(arg, int) for arg in evaluated)


def op_is_list(seval, args):
    '''Returns true if all arguments evaluate to lists'''
    evaluated = seval.eval_args(args)
    return all(isinstance(arg, (WList, list)) for arg in evaluated)


def op_convert_binary(seval, args):
    '''Converts an integer to a binray string of width bits'''
    assert len(args) == 1 or len(args) == 2, 'convert/bin: expects at least one argument (convert/bin expr:int width:int)'
    evaluated = seval.eval_args(args)
    value = evaluated[0]
    width = evaluated[1] if len(args) == 2 else 0
    assert isinstance(value, int), 'convert/bin: first argument must evaluate to int'
    assert isinstance(width, int), 'convert/bin: second argument must evaluate to int'
    return f'{value:0{width}b}'.format(value=value, width=width) # pylint: disable=C0116


def op_string_to_int(seval, args):
    '''Converts a string to an integer '''
    assert len(args) == 1 or len(args) == 2, 'string->int: expects exactly one argument (string->int expr:str base:int?)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, str), 'string->int: argument must evaluate to srting'
    if len(args) == 1:
        return int(evaluated)

    assert isinstance(args[1], int), 'string->int: argument base must be integer'
    assert args[1] in [2, 8, 10, 16] , 'string->int: valid base values (2 8 10 16)'
    return int(evaluated, args[1])


def op_bits_to_sint(seval, args):
    '''Converts a string to a signed integer '''
    assert len(args) == 1, 'bits->sint: expects exactly one argument (bits->sint expr:str)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, str), 'bits->sint: argument must evaluate to srting'

    if evaluated[0] == '1':
        u_res = int(''.join('1' if x == '0' else '0' for x in evaluated), 2) + 1
        return -u_res

    return int(evaluated, 2)


def op_symbol_to_string(seval, args):
    '''Converts a symbol a to a string "a" '''
    assert len(args) == 1, 'symbol->string: expects exactly one argument (symbol->string expr:symbol)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, Symbol), 'symbol->string: argument must evaluate to symbol'
    return evaluated.name


def op_string_to_symbol(seval, args):
    '''Converts a string to a symbol'''
    assert len(args) == 1, 'string->symbol: expects exactly one argument (string->symbol expr:symbol)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, str), 'symbol->string: argument must evaluate to symbol'
    return Symbol(evaluated)


def op_int_to_string(seval, args):
    '''Converts an integer to its string representation'''
    assert len(args) == 1, 'int->string: expects exactly one argument (symbol->string expr:symbol)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, int), 'int->string: argument must evaluate to symbol'
    return str(evaluated)


type_operators = {
    Operator.IS_DEFINED.value: op_is_defined,
    Operator.IS_ATOM.value: op_is_atom,
    Operator.IS_SYMBOL.value: op_is_symbol,
    Operator.IS_STRING.value: op_is_string,
    Operator.IS_INT.value: op_is_int,
    Operator.IS_LIST.value: op_is_list,
    Operator.CONVERT_BINARY.value: op_convert_binary,
    Operator.STRING_TO_INT.value: op_string_to_int,
    Operator.BITS_TO_SINT.value: op_bits_to_sint,
    Operator.STRING_TO_SYMBOL.value: op_string_to_symbol,
    Operator.SYMBOL_TO_STRING.value: op_symbol_to_string,
    Operator.INT_TO_STRING.value: op_int_to_string,
}



'''Test wal trace readers'''
import unittest

from wal.core import Wal
from wal.ast_defs import WalEvalError

class BasicParserTest(unittest.TestCase):
    '''Test trace readers'''

    def test_sv_arrays(self):
        '''Test SystemVerilog arrays'''
        for trace_format in ['vcd', 'fst']:
            wal = Wal()
            wal.load(f'tests/traces/sv_struct_array.{trace_format}')
            self.assertEqual(wal.eval_str('TOP.tb.data<0>.data.something'), 5)


class ReadUndefinedSignalTest(unittest.TestCase):
    '''Test trace readers'''

    def test_read_undefined_signals(self):
        '''Test SystemVerilog arrays'''
        for trace_format in ['vcd', 'fst']:
            wal = Wal()
            wal.load(f'tests/traces/sv_struct_array.{trace_format}')

            with self.assertRaises(WalEvalError):
                wal.eval_str('TOP.tb.data<0>.wrongname')

            with self.assertRaises(WalEvalError):
                wal.eval_str('(in-scope "TOP.tb.data<0>" ~wrongname)')

            with self.assertRaises(WalEvalError):
                wal.eval_str('(in-group "TOP.tb.data<0>" #.wrongname)')



'''Parsers for WAWK'''
# pylint: disable=C0103,C0116
from lark import Lark, Transformer
from wawk.ast_defs import Statement
from wal.ast_defs import Symbol as S
from wal.ast_defs import Operator as Op
from wal.reader import operators

WAWK_GRAMMAR = r"""
    ?expr: symbol
         | fcall
         | array_get
         | "(" expr ")"
         | string
         | list
         | neg
         | sum_s
         | or_s
         | SIGNED_INT -> number
         | INT -> number

    line_expr: assign
         | array_set

    block_expr: forstmt
         | cond
         | in_scope
         | in_group

    substatement: (subline | block_expr | block) ";"*
    subline: (line_expr | expr) ";"

    list : "[" [expr ("," expr)*] "]"

    forstmt: forin | forinarray | forvar
    forin: "for" "(" base_symbol "in" (list | fcall | base_symbol) ")" substatement
    forinarray: "for" "(" base_symbol "," base_symbol "in" (fcall | base_symbol) ")" substatement
    forvar: "for" "(" assign_std ";" expr ";" expr  ")" substatement

    cond : ifstmt
    ifstmt : "if" "(" expr ")" substatement ("else" substatement)?

    symbol : simple_symbol | sliced_symbol | bit_symbol
    simple_symbol : base_symbol | scoped_symbol | grouped_symbol | timed_symbol | bit_symbol | sliced_symbol
    scoped_symbol : "~" base_symbol
    grouped_symbol : "#" base_symbol
    timed_symbol : base_symbol "@" SIGNED_INT
    !base_symbol : (LETTER | "_") (LETTER | INT | "_" | "$" | ".")*
    bit_symbol : simple_symbol "[" INT "]"
    sliced_symbol : simple_symbol "[" INT ":" INT "]"

    neg.6 : a_neg | expr
    a_neg.6 : u_op expr

    sum_s.5: a_sum_s | mul
    a_sum_s.5: sum_s a_s_op sum_s
    mul.4: a_mul | expr
    a_mul.4: mul m_d_op mul

    or_s.3: a_or_s | and_s
    a_or_s.3: or_s or_op or_s
    and_s.2: a_and_s | comp
    a_and_s.2: and_s and_op and_s
    comp.1: a_comp | expr
    a_comp.1: and_s comp_op and_s

    !u_op : "!"
    !m_d_op : "*" | "/"
    !a_s_op : "+" | "-"
    !comp_op : "==" | "!=" | ">" | "<" | ">=" | "<="
    !and_op : "&&"
    !or_op : "||"

    block : "{" (substatement)* "}"
    assign : assign_arith | assign_std | define
    assign_std : base_symbol "=" expr
    define : "var" base_symbol "=" expr
    !assign_arith : base_symbol ("+"| "-" | "*" | "/") "=" expr

    function : "function" base_symbol "(" [base_symbol ("," base_symbol)*] ")" block
    fcall : base_symbol "(" [expr ("," expr)*] ")"

    array_op : array_set | array_get
    array_get : base_symbol "[" (INT | string | symbol) ("," (INT | string | symbol))* "]"
    array_set : base_symbol "[" (INT | string | symbol) ("," (INT | string | symbol))* "]" "=" expr

    in_scope : "scope" "(" (symbol | list | fcall) ")" substatement
    in_group : "group" "(" (list | fcall | expr) ")" substatement

    _NL: /(\r?\n)+\s*/
    statement : [expr ("," expr)* ":"] block
    line : function | statement
    program : [line (_NL line)*]

    COMMENT: /\/\/[^\n]*/

    string : ESCAPED_STRING

    %import common.ESCAPED_STRING
    %import common.SIGNED_INT
    %import common.INT
    %import common.WORD
    %import common.WS_INLINE
    %import common.WS
    %import common.LETTER
    %import common.NEWLINE
    %ignore WS
    %ignore COMMENT
    """

class TreeToWal(Transformer):
    '''Transforms a parsed tree into WAL data structures'''

    def function(self, f):
        args = f[1:-1] if f[1] else []
        return Statement([S('BEGIN')], [S('defun'), f[0], args, f[-1]])

    statement = lambda self, s: Statement(s[:-1], s[-1])
    line = lambda self, line: line[0]
    program = lambda self, p: p

    substatement = lambda self, s: s[0]
    subline = lambda self, line: line[0]
    block_expr = lambda self, b: b[0]
    line_expr = lambda self, line: line[0]

    string = lambda self, s: s[0][1:-1].encode('utf-8').decode('unicode_escape')
    number = lambda self, n: int(n[0])
    simple_symbol = lambda self, s: s[0]
    symbol = lambda self, s: s[0]
    base_symbol = lambda self, s: S(''.join(list(map(lambda x: x.value, s))))
    scoped_symbol = lambda self, s: [Op.SCOPED, s[0]]
    grouped_symbol = lambda self, s: [Op.RESOLVE_GROUP, s[0]]
    timed_symbol = lambda self, s: [Op.REL_EVAL, s[0], self.number([s[1].value])]
    bit_symbol = lambda self, s: [Op.SLICE, s[0], self.number([s[1].value])]
    sliced_symbol = lambda self, s: [Op.SLICE, s[0], self.number([s[1].value]), self.number([s[2].value])]
    block = lambda self, b: [Op.DO] + b
    assign = lambda self, s: s[0]
    assign_std = lambda self, a: [Op.SET, [a[0], a[1]]]
    define = lambda self, a: [Op.DEFINE, a[0], a[1]]
    assign_arith = lambda self, a: [Op.SET, [a[0], [Op(a[1]), a[0], a[3]]]]
    forin = lambda self, f: [Op.MAP, [Op.FN, [f[0]], f[2]], f[1]]
    forinarray = lambda self, f: [Op.MAPA, [Op.FN, [f[0], f[1]], f[3]], f[2]]
    forvar = lambda self, f: [Op.LET, [f[0][1]], [Op.WHILE, f[1], [Op.DO, f[3], f[2]]]]

    neg = lambda self, x: x[0]
    a_neg = lambda self, x: [Op(x[0].children[0].value), x[1]]
    mul = lambda self, x: x[0]
    a_mul = lambda self, x: [Op(x[1].children[0].value), x[0], x[2]]
    sum_s = lambda self, x: x[0]
    a_sum_s = lambda self, x: [Op(x[1].children[0].value), x[0], x[2]]
    and_s = lambda self, x: x[0]
    a_and_s = lambda self, x: [Op(x[1].children[0].value), x[0], x[2]]
    or_s = lambda self, x: x[0]
    a_or_s = lambda self, x: [Op(x[1].children[0].value), x[0], x[2]]
    comp = lambda self, x: x[0]

    array_op = lambda self, a: a[0]

    def array_get(self, a):
        array = a[0]
        indices = a[1:]
        return [S('geta/default'), array, 0, [Op.ADD, *indices]]

    def array_set(self, a):
        array = a[0]
        indices = a[1:-1]
        value = a[-1]
        return [Op.SETA, array, [Op.ADD, *indices], value]

    cond = lambda self, c: c[0]
    ifstmt = lambda self, c: [Op.IF] + c
    in_scope = lambda self, s: [Op.SCOPED, s[0], s[1]]
    in_group = lambda self, g: [Op.IN_GROUPS, g[0], g[1]]

    def list(self, ls):
        return [Op.LIST] + (ls if ls != [None] else [])

    forstmt = lambda self, f: f[0]
    null = lambda self, _: None
    true = lambda self, _: True
    false = lambda self, _: False

    def fcall(self, f):
        args = f[1:] if f[1:] != [None] else []
        if f[0].name in operators:
            expr = [Op(f[0].name)] + args
        else:
            expr = [f[0]] + args

        return expr

    def a_comp(self, x):
        if x[1].children[0].value == '==':
            return [Op.EQ, x[0], x[2]]

        return [Op(x[1].children[0].value), x[0], x[2]]

def parse_wawk(code):
    json_parser = Lark(WAWK_GRAMMAR, start='program')
    parsed = json_parser.parse(code)
    return TreeToWal().transform(parsed)

def test():
    json_parser = Lark(WAWK_GRAMMAR, start='expr')
    parsed = json_parser.parse('group (groups("ready", "valid")) {x}')
    print(TreeToWal().transform(parsed))



'''Module containing the AST definitions'''
from dataclasses import dataclass
from enum import Enum, unique
from collections import UserList

@unique
class Operator(Enum):
    '''Enum for built in operations'''
    # waveform
    LOAD = 'load'
    UNLOAD = 'unload'
    STEP = 'step'
    REPL = 'repl'
    LOADED_TRACES = 'loaded-traces'
    IS_SIGNAL = 'signal?'
    # basic
    REQUIRE = 'require'
    EVAL_FILE = 'eval-file'
    # maths
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    EXP = '**'
    FLOOR = 'floor'
    CEIL = 'ceil'
    ROUND = 'round'
    MOD = 'mod'
    # bitwise
    BOR = 'bor'
    BAND = 'band'
    BXOR = 'bxor'
    # logic
    NOT = '!'
    EQ = '='
    NEQ = '!='
    LARGER = '>'
    SMALLER = '<'
    LARGER_EQUAL = '>='
    SMALLER_EQUAL = '<='
    AND = '&&'
    OR = '||'
    PRINT = 'print'
    PRINTF = 'printf'
    SET = 'set'
    DEFINE = 'define'
    LET = 'let'
    IF = 'if'
    CASE = 'case'
    WHILE = 'while'
    DO = 'do'
    ALIAS = 'alias'
    UNALIAS = 'unalias'
    QUOTE = 'quote'
    QUASIQUOTE = 'quasiquote'
    UNQUOTE = 'unquote'
    EVAL = 'eval'
    PARSE = 'parse'
    DEFMACRO = 'defmacro'
    MACROEXPAND = 'macroexpand'
    GENSYM = 'gensym'
    FN = 'fn'
    GET = 'get'
    CALL = 'call'
    IMPORT = 'import'
    LIST = 'list'
    # FOR = 'for'
    FIRST = 'first'
    SECOND = 'second'
    LAST = 'last'
    REST = 'rest'
    IN = 'in'
    MAP = 'map'
    MAX = 'max'
    MIN = 'min'
    FOLD = 'fold'
    LENGTH = 'length'
    AVERAGE = 'average'
    ZIP = 'zip'
    RANGE = 'range'
    TYPE = 'type'
    REL_EVAL = 'reval'
    ARRAY = 'array'
    SETA = 'seta'
    GETA = 'geta'
    DELA = 'dela'
    MAPA = 'mapa'
    ALLSCOPES = 'all-scopes'
    SCOPED = 'in-scope'
    RESOLVE_SCOPE = 'resolve-scope'
    SETSCOPE = 'set-scope'
    UNSETSCOPE = 'unset-scope'
    GROUPS = 'groups'
    IN_GROUP = 'in-group'
    IN_GROUPS = 'in-groups'
    RESOLVE_GROUP = 'resolve-group'
    SLICE = 'slice'
    # types
    IS_DEFINED = 'defined?'
    IS_ATOM = 'atom?'
    IS_SYMBOL = 'symbol?'
    IS_STRING = 'string?'
    IS_INT = 'int?'
    IS_LIST = 'list?'
    CONVERT_BINARY = 'convert/bin'
    STRING_TO_INT = 'string->int'
    BITS_TO_SINT = 'bits->sint'
    STRING_TO_SYMBOL = 'string->symbol'
    SYMBOL_TO_STRING = 'symbol->string'
    INT_TO_STRING = 'int->string'
    # special
    FIND = 'find'
    FIND_G = 'find/g'
    WHENEVER = 'whenever'
    FOLD_SIGNAL = 'fold/signal'
    SIGNAL_WIDTH = 'signal-width'
    SAMPLE_AT = 'sample-at'
    TRIM_TRACE = 'trim-trace'
    # system
    EXIT = 'exit'
    # virtual signals
    DEFSIG = 'defsig'
    NEWTRACE = 'new-trace'
    DUMPTRACE = 'dump-trace'


operators = set([op.value for op in Operator])


class WList(UserList):

    def __init__(self, data, line_info=None):
        super().__init__(data)
        self.line_info = line_info

    def __str__(self):
        txt = super().__str__()
        return f'WList({txt}, {self.line_info})'

    def __eq__(self, other):
        if isinstance(other, WList):
            return list.__eq__(self.data, other.data)

        return list.__eq__(self.data, other)

    def strip_line_info(self):
        return [d.strip_line_info() if isinstance(d, WList) else d for d in self.data]


class Symbol:
    '''Symbol class'''

    def __init__(self, name, steps=None, line_info=('', 0, 0)):
        self.name = name
        self.steps = steps
        self.line_info = line_info

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.name == other.name and self.steps == other.steps

        return False


class UserOperator:
    '''Class that wraps a user specified operator'''

    def __init__(self, name):
        assert name not in operators, f'redefining {name} is not allowed'
        self.name = name

    def __repr__(self):
        return self.name


class Environment:
    '''Object that holds a WAL environment '''

    def __init__(self, parent=None):
        self.environment = {}
        self.parent = parent

    def define(self, name, value):
        '''Define new variable in this context '''
        assert name not in self.environment, f'variable {name} already defined'
        self.environment[name] = value

    def undefine(self, name):
        '''Remove definition from this context'''
        assert name in self.environment, f'variable {name} is not defined'
        del self.environment[name]

    def is_defined(self, name):
        '''Check if name is defined somewhere and return that environment'''

        if name in self.environment:
            return self.environment

        if self.parent:
            return self.parent.is_defined(name)

        return False

    def write(self, name, value):
        '''Write to variable name'''

        if name in self.environment:
            self.environment[name] = value
        else:
            assert self.parent
            self.parent.write(name, value)

    def read(self, name):
        '''Read from variable name'''

        if name in self.environment:
            return self.environment[name]

        assert self.parent, f'variable {name} is undefined'
        return self.parent.read(name)



class Closure:
    '''Class implementing a closure '''

    def __init__(self, environment, args, expression, name='lambda'):
        self.name = name
        self.environment = environment
        self.args = args
        self.expression = expression


class Macro:
    '''Class that holds a macro'''

    def __init__(self, name, args, expression):
        self.name = name
        self.args = args
        self.expression = expression


@dataclass
class Unquote:
    '''Utility class for unquote syntax'''
    content: any


@dataclass
class UnquoteSplice:
    '''Utility class for unquote splice syntax'''
    content: any


class VirtualSignal:
    '''Holds information about a virtual signal'''
    name: str

    def __init__(self, name, expr, trace, seval, width=32): # pylint: disable=R0913
        self.name = name
        self.expr = expr
        self.width = width
        self.trace = trace
        self.seval = seval
        self.dependencies = []
        self.cache = {}

    @property
    def value(self):
        '''Returns the value of this virtual signal'''
        index = self.trace.index
        ts = self.trace.timestamps[index]
        if ts in self.cache:
            res = self.cache[ts]
        else:
            res = self.seval.eval_args(self.expr)[-1]
            self.cache[ts] = res

        return res


class WalEvalError(Exception):

    def __init__(self, message=''):
        self.trace = []
        self.message = message

    def add(self, f):
        self.trace.append(f)

    def print(self):
        if self.trace:
            print()
            print("Backtrace:")
            for function in reversed(self.trace):
                if isinstance(function, Symbol):
                    line = function.line_info['line']
                    filename = function.line_info['filename']
                    print(f'{function}, line {line} in {filename}')
                else:
                    print(function)



'''Trace implementation for the VCD file format '''
import re
import sys

from wal.trace.trace import Trace

class TraceVcd(Trace):
    '''Holds data for one vcd trace.'''

    SKIPPED_COMMANDS_HEADER = set(['$comment', '$version', '$date'])

    def __init__(self, filename, tid, container, from_string=False, keep_signals=None):
        super().__init__(tid, filename, container)
        self.timestamps = []
        self.lookup = None # performs index translation after set-sampling
        self.scopes = []
        self.rawsignals = []
        self.all_ids = set()
        self.index2ts = []
        self.name2id = {}
        self.data = {}
        self.signalinfo = {}
        self.filename = filename
        self.keep_signals = set(keep_signals) if keep_signals else None
        if from_string:
            self.parse(filename)
        else:
            try:
                with open(filename) as f:
                    self.parse(f.read())
            except FileNotFoundError:
                print(f'Error while loading {filename}. File not found.')
                sys.exit(1)


        self.all_timestamps = self.timestamps.copy()
        self.index = 0
        self.max_index = len(self.index2ts) - 1
        self.signals = set(Trace.SPECIAL_SIGNALS + self.rawsignals)

        self.id2name = {v: k for k, v in self.name2id.items()}
        self.rawsignals_by_handle = [self.id2name[s] for s in self.all_ids]
        self.signals_by_handle = set(self.rawsignals_by_handle)

    def parse(self, vcddata):
        scope = []
        tokens = vcddata.split()

        i = 0
        header_done = False
        # header section
        while (not header_done) and tokens:
            if tokens[i] == '$scope':
                name = tokens[i + 2]

                # array entries should not clash with WAL operators
                name = re.sub(r'\[([0-9]+)\]', r'<\1>', name)
                name = re.sub(r'\(([0-9]+)\)', r'<\1>', name)

                scope.append(name)
                self.scopes.append('.'.join(scope))
                i += 4
            elif tokens[i] == '$var':
                kind = tokens[i + 1]
                width = tokens[i + 2]
                id = tokens[i + 3]
                name = tokens[i + 4]

                # remove slice info from names
                name = re.sub(r'\[[0-9]+:[0-9]+\]', '', name)
                # array signals should not clash with WAL operators
                name = re.sub(r'\[([0-9]+)\]', r'<\1>', name)
                name = re.sub(r'\(([0-9]+)\)', r'<\1>', name)

                # only append scope. if not in root scope
                if scope:
                    fullname = '.'.join(scope) + '.' + name
                else:
                    fullname = name
                
                if not self.keep_signals or (fullname in self.keep_signals):
                    self.all_ids.add(id)
                    self.rawsignals.append(fullname)
                    self.name2id[fullname] = id
                    self.signalinfo[id] = {
                        'id': id,
                        'name': fullname,
                        'width': int(width),
                        'kind': kind,
                        'data': {}
                    }

                if tokens[i + 5] == '$end':
                    i += 6
                elif tokens[i + 5][0] == '[':
                    i += 7
                else:
                    assert False, 'VCD error'
            elif tokens[i] == '$upscope':
                scope.pop()
                i += 2
            elif tokens[i] == '$enddefinitions':
                i += 2
                header_done = True
            elif tokens[i] == '$timescale':
                if tokens[i + 3] == '$end':
                    self.timescale = tokens[i + 1] + tokens[i + 2]
                    i += 4
                elif tokens[i + 2] == '$end':
                    self.timescale = tokens[i + 1]
                    i += 3
            elif tokens[i] in TraceVcd.SKIPPED_COMMANDS_HEADER:
                while tokens[i] != '$end':
                    i += 1

                i += 1
            else:
                # this should not happen
                i += 1

        # parse dump section
        time = 0
        n_tokens = len(tokens)
        # fill initially with all Xs
        self.data = {id: ['x'] for id in self.all_ids}
        SCALARS = ['x', 'z', 'X', 'Z']

        while i < n_tokens:
            first_char = tokens[i][0]
            if first_char == '0' or first_char == '1':
                id = tokens[i][1:]
                if id in self.all_ids:
                    self.data[id][-1] = 'b' + first_char

                i += 1
            elif first_char == 'b' or first_char == 'r':
                # n-bit vector of format b0000 id
                id = tokens[i + 1]
                if id in self.all_ids:
                    self.data[id][-1] = tokens[i]

                i += 2
            elif first_char == '#':
                time = int(tokens[i][1:])
                i += 1
                # copy old values
                for id in self.all_ids:
                    self.data[id].append(self.data[id][-1])

                self.timestamps.append(time)
                self.index2ts.append(time)
            elif first_char in SCALARS:
                # scalar value change
                id = tokens[i][1:]
                if id in self.all_ids:
                    self.data[id][-1] = tokens[i][0]

                i += 1
            elif tokens[i] == '$comment':
                while tokens[i] != '$end':
                    i += 1

                i += 1
            else:
                # token is most likely one of ['$dumpvars', '$dumpall', '$dumpoff', '$dumpon', '$end']
                # we skip these commands and just read the following changes
                i += 1                

        # remove inital values
        for id in self.all_ids:
            self.data[id].pop(0)

        # modify data to be a lookup by signal name, removes the indirection via the id
        data_by_name = {signal: self.data[self.name2id[signal]] for signal in self.rawsignals}
        self.data = data_by_name

    def set_sampling_points(self, new_indices):
        '''Updates the indices at which data is sampled'''
        self.lookup = dict(enumerate(new_indices))
        new_timestamps = [self.all_timestamps[i] for i in new_indices]
        self.timestamps = list(dict.fromkeys(new_timestamps))
        self.timestamps = dict(enumerate(self.timestamps))
        # stores current time stamp
        self.index = 0
        self.max_index = len(self.timestamps.keys()) - 1

    def access_signal_data(self, name, index):
        if self.lookup:
            value = self.data[name][self.lookup[index]]
        else:
            value = self.data[name][index]

        if value[0] == 'b':
            return int(value[1:], 2)
        elif value[0] == 'r':
            return float(value[1:])
        else:
            return value
    
    def signal_width(self, name):
        '''Returns the width of a signal'''
        return self.signalinfo[self.name2id[name]]['width']



'''Test equivalence of trace loaders'''
import unittest
from itertools import combinations

from wal.core import Wal


class TraceEqTest():

    def setUp(self):
        self.traces = map(lambda f: f'tests/traces/{f}', self.traces)

    def wal_eval(self, data):
        wal = Wal()
        wal.load(data[0])
        return wal.eval_str(data[1])

    def eval_eq(self, p):
        res = list(map(lambda t: self.wal_eval([t, p]), self.traces))

        for pair in combinations(res, 2):
            self.assertEqual(pair[0], pair[1])

    def test_trace_name(self):
        self.eval_eq('TRACE-NAME')

    def test_signals(self):
        self.eval_eq('SIGNALS')

    def test_max_index(self):
        self.eval_eq('MAX-INDEX')

    def test_scopes(self):
        self.eval_eq('SCOPES')

    def test_goto_end(self):
        self.eval_eq("(do (while (step) 0) INDEX)")

    def test_ts_list(self):
        self.eval_eq("(map (fn [x] TS@x) (range MAX-INDEX))")

    def test_all_signal_values(self):
        self.eval_eq("(map (fn [x] (map get SIGNALS)@x) (range MAX-INDEX))")


class CounterEqualTest(TraceEqTest, unittest.TestCase):
    '''Test counter traces for equality'''

    def setUp(self):
        self.traces = ['counter.vcd', 'counter.fst']
        super().setUp()

    def test_group_clk(self):
        self.eval_eq('(groups clk)')

    def test_rising_clocks(self):
        self.eval_eq('(find (rising tb.clk))')

    def test_overflow(self):
        self.eval_eq('(count tb.overflow')

    def test_signals(self):
        self.eval_eq('SIGNALS')

    def test_scopes(self):
        self.eval_eq('SCOPES')

    def test_local_scopes(self):
        self.eval_eq('LOCAL-SCOPES')
        self.eval_eq("(in-scope 'tb CS)")
        self.eval_eq("(in-scope 'tb LOCAL-SCOPES)")

    def test_local_signals(self):
        wal = "(in-scope 'tb LOCAL-SIGNALS)"
        res = list(map(lambda t: self.wal_eval([t, wal]), self.traces))

        for pair in combinations(res, 2):
            self.assertEqual(sorted(pair[0]), sorted(pair[1]))


class SVArrayEqualTest(TraceEqTest, unittest.TestCase):
    '''Test counter traces for equality'''

    def setUp(self):
        self.traces = ['sv_struct_array.vcd', 'sv_struct_array.fst']
        super().setUp()

    def test_group_clk(self):
        self.eval_eq('(groups something)')

    def test_signals(self):
        self.eval_eq('SIGNALS')

    def test_scopes(self):
        self.eval_eq('SCOPES')

    def test_local_scopes(self):
        self.eval_eq('LOCAL-SCOPES')
        self.eval_eq("(in-scope 'tb CS)")
        self.eval_eq("(in-scope 'tb LOCAL-SCOPES)")

    def test_local_signals(self):
        wal = "(in-scope 'tb LOCAL-SIGNALS)"
        res = list(map(lambda t: self.wal_eval([t, wal]), self.traces))

        for pair in combinations(res, 2):
            self.assertEqual(sorted(pair[0]), sorted(pair[1]))



'''Test wal list eval logic'''
from wal.ast_defs import Operator as Op
from wal.ast_defs import WalEvalError
from .test_eval import OpTest

# pylint: disable=C0103
# pylint: disable=W0201


class BasicListTest(OpTest):
    '''Test basic list functions such as list creation'''

    def test_list_construcion(self):
        '''Test constructing lists'''
        self.checkEqual('(list)', [])
        self.checkEqual('(list 1 2)', [1, 2])
        self.checkEqual('(list (list 1 2) 3)', [[1, 2], 3])

    def test_first_second_last(self):
        '''test accessing list elements'''
        l1 = "'(1 2 3)"
        l2 = "'((1 2) 3)"
        self.checkEqual(f'(first {l1})', 1)
        self.checkEqual(f'(first {l2})', [1, 2])

        self.checkEqual(f'(second {l1})', 2)
        self.checkEqual(f'(second {l2})', 3)

        self.checkEqual(f'(last {l1})', 3)
        self.checkEqual(f'(last {l2})', 3)

        self.checkEqual("(first '(1))", self.w.eval_str("(last '(1))"))
        self.checkEqual("(second '(1 2))", self.w.eval_str("(last '(1 2))"))

        for op in [Op.FIRST, Op.SECOND, Op.LAST]:
            with self.assertRaises(WalEvalError):
                self.w.eval_str(f'({op.value})')
            with self.assertRaises(WalEvalError):
                self.w.eval_str(f'({op.value} 1 2)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str("(second '(1))")

    def test_rest(self):
        '''test rest operator'''
        self.checkEqual("(rest '(1 2 3))", [2, 3])
        self.checkEqual("(rest '(1 2))", [2])
        self.checkEqual("(rest '(1))", [])
        self.checkEqual("(rest '())", [])

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(rest)')

    def test_in(self):
        '''test in operator'''
        l1 = "'(1 2 3 4)"
        self.checkEqual(f'(in 1 {l1})', True)
        self.checkEqual(f'(in 1 2 {l1})', True)
        self.checkEqual(f'(in 0 {l1})', False)
        self.checkEqual(f'(in 1 0 {l1})', False)

        self.checkEqual(f'(in "a" {l1})', False)

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(in)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(in 1)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(in 1 2)')

    def test_map(self):
        '''test map operator'''
        l1 = "'(1 2 3 4)"
        l2 = "'(\"a\" \"b\")"
        f = '(fn (x) (+ x 1))'
        self.checkEqual(f'(map {f} {l1})', [2, 3, 4, 5])
        self.checkEqual(f'(map {f} {l2})', ['a1', 'b1'])

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(map)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(map f)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str(f'(map {l1})')
        with self.assertRaises(WalEvalError):
            self.w.eval_str(f'(map {l1} {f})')
        with self.assertRaises(WalEvalError):
            self.w.eval_str(f'(map {l1} {l2})')
        with self.assertRaises(WalEvalError):
            self.w.eval_str(f'(map {f} {l1} {l2})')



'''Implementations for type related functions such as type checking'''
from wal.ast_defs import Operator, Symbol, WList
from wal.trace.virtual import TraceVirtual


def op_defsig(seval, args):
    '''Creates a new signal whose value is calculated by evaluating
    the body expressions. '''
    assert len(args) > 1, 'virtual: expects at least two arguments (virtual name body+)'
    scope_name = seval.global_environment.read('CS')
    scope_sep = '.' if scope_name != '' else ''
    scope = scope_name + scope_sep
    group = seval.global_environment.read('CG') 
    if group != '':
        # groups already include the scope so reset it
        scope = ''

    name = f'{scope}{group}{args[0].name}'

    def resolve(expr):
        if isinstance(expr, WList):
            if len(expr) == 2 and isinstance(expr[1], Symbol):
                if expr[0] == Operator.RESOLVE_SCOPE:
                    return Symbol(f'{scope}{expr[1].name}')
                elif expr[0] == Operator.RESOLVE_GROUP:
                    return Symbol(f'{group}{expr[1].name}')

            return WList(list(map(resolve, expr)))

        return expr

    body = WList(list(map(resolve, args[1:])))
    seval.traces.add_virtual_signal(name, body, seval)


def op_new_trace(seval, args):
    '''Create a new virtual trace for id'''
    assert len(args) == 2, 'new-trace: expects one argument (new-trace id:symbol max-index:int)'
    assert isinstance(args[0], Symbol), 'new-trace: first argument must be a symbol'
    assert isinstance(args[1], int), 'new-trace: second argument must be an int'

    seval.traces.traces[args[0].name] = TraceVirtual(args[0].name, args[1], seval.traces)
    seval.traces.n_traces += 1


def op_dump_trace(seval, args):
    '''Dumps the given trace as in vcd format.'''
    seval.traces.traces[args[0].name].dump_vcd()


virtual_operators = {
    Operator.DEFSIG.value: op_defsig,
    Operator.NEWTRACE.value: op_new_trace,
    Operator.DUMPTRACE.value: op_dump_trace,
}



'''Test resoling variables logic'''
import unittest

from wal.core import Wal
from wal.reader import read_wal_sexpr
from wal.passes import resolve
from wal.ast_defs import Symbol as S
from wal.ast_defs import Operator as Op
from wal.ast_defs import WalEvalError


class BasicResolveTest(unittest.TestCase):
    '''Test built-in functions'''

    def setUp(self):
        self.wal = Wal()

    def checkEqual(self, txt, res):
        '''eval first argument and check if result matches second argument '''
        self.assertEqual(resolve(read_wal_sexpr(txt)), res)

    def checkRaises(self, txt):
        '''eval first argument and check if assertion is raised'''
        self.assertRaises(WalEvalError, lambda code: resolve(read_wal_sexpr(code)), txt)

    def test_resolve_num(self):
        '''Test resolution of numbers'''

        self.checkEqual('1', 1)
        self.checkEqual('-5', -5)
        self.checkEqual('1.2', 1.2)
        self.checkEqual('0xff123', 0xff123)

    def test_use_afeter_define(self):
        self.checkEqual('(do (define x 5) x)', [Op.DO, [Op.DEFINE, S('x'), 5], S('x', 0)])
        self.checkEqual('(do (define x 5) (define y 2) y x)', [Op.DO, [Op.DEFINE, S('x'), 5], [Op.DEFINE, S('y'), 2], S('y', 0), S('x', 0)])

    def test_use_in_let(self):
        self.checkEqual('(let ([x 5]) x)', [Op.LET, [[S('x'), 5]], S('x', 0)])
        self.checkEqual('(let ([x 5] [y 2]) x)', [Op.LET, [[S('x'), 5], [S('y'), 2]], S('x', 0)])
        self.checkEqual('(let ([x 5] [y 2]) y x)', [Op.LET, [[S('x'), 5], [S('y'), 2]], S('y', 0), S('x', 0)])
        self.checkEqual('(let ([x 5] [y 2]) (+ y x))', [Op.LET, [[S('x'), 5], [S('y'), 2]], [Op.ADD, S('y', 0), S('x', 0)]])
        self.checkEqual('(let ([x 5]) (let ([y 2]) (+ y x)))', [Op.LET, [[S('x'), 5]], [Op.LET, [[S('y'), 2]], [Op.ADD, S('y', 0), S('x', 1)]]])

    def test_use_after_let(self):
        self.checkEqual('(do (let ([x 5]) x) x)', [Op.DO, [Op.LET, [[S('x'), 5]], S('x', 0)], S('x')])
        self.checkEqual('(do (define y 2) (let ([x 5]) y))', [Op.DO, [Op.DEFINE, S('y'), 2], [Op.LET, [[S('x'), 5]], S('y', 1)]])

    def test_shadow_in_let(self):
        self.checkEqual('(do (define x 2) (let ([x 5]) x) x)', [Op.DO, [Op.DEFINE, S('x'), 2], [Op.LET, [[S('x'), 5]], S('x', 0)], S('x', 0)])

    def test_undefined_symbol_should_stay_symbol(self):
        self.checkEqual('x', S('x'))
        self.checkEqual('(do x)', [Op.DO, S('x')])
        self.checkEqual('(do 5 x)', [Op.DO, 5, S('x')])
        self.checkEqual('(define x x)', [Op.DEFINE, S('x'), S('x')])

    def test_lambda_resolution(self):
        self.checkEqual('(fn [x] x)', [Op.FN, [S('x')], S('x', 0)])
        self.checkEqual('(fn [x] y)', [Op.FN, [S('x')], S('y')])
        self.checkEqual('(do (define y 5) (fn [x] y))', [Op.DO, [Op.DEFINE, S('y'), 5], [Op.FN, [S('x')], S('y', 1)]])
        self.checkEqual('(fn [x] (fn [y] x))', [Op.FN, [S('x')], [Op.FN, [S('y')], S('x', 1)]])
        self.checkEqual('(fn [x] (fn [y] y))', [Op.FN, [S('x')], [Op.FN, [S('y')], S('y', 0)]])
        self.checkEqual('(fn [x y] (fn [y] y))', [Op.FN, [S('x'), S('y')], [Op.FN, [S('y')], S('y', 0)]])
        self.checkEqual('(fn [x y] (fn [z] y))', [Op.FN, [S('x'), S('y')], [Op.FN, [S('z')], S('y', 1)]])

    def test_resolution_in_set(self):
        self.checkEqual('(set [x 5])', [Op.SET, [S('x'), 5]])
        self.checkEqual('(do (define x 0) (set [x 5]))', [Op.DO, [Op.DEFINE, S('x'), 0], [Op.SET, [S('x', 0), 5]]])
        self.checkEqual('(do (define x 0) (set [x x]))', [Op.DO, [Op.DEFINE, S('x'), 0], [Op.SET, [S('x', 0), S('x', 0)]]])
        self.checkEqual('(do (define x 0) (set [x y]))', [Op.DO, [Op.DEFINE, S('x'), 0], [Op.SET, [S('x', 0), S('y')]]])

class ResolveEvalTest(unittest.TestCase):
    '''Test built-in functions'''

    def setUp(self):
        self.wal = Wal()

    def checkEqual(self, txt, res):
        '''eval first argument and check if result matches second argument '''
        self.assertEqual(self.wal.run_str(txt), res)

    def checkRaises(self, txt):
        '''eval first argument and check if assertion is raised'''
        self.assertRaises(WalEvalError, lambda code: self.wal.run_str(code), txt)

    def test_resolve_num(self):
        '''Test resolution of numbers'''

        self.checkEqual('1', 1)
        self.checkEqual('-5', -5)
        self.checkEqual('1.2', 1.2)
        self.checkEqual('0xff123', 0xff123)

    def test_use_afeter_define(self):
        self.checkEqual('(do (define x 5) x)', 5)
        self.checkEqual('(do (define x 5) (define y 2) y x)', 5)

    def test_use_in_let(self):
        self.checkEqual('(let ([x 5]) x)', 5)
        self.checkEqual('(let ([x 5] [y 2]) x)', 5)
        self.checkEqual('(let ([x 5] [y 2]) y x)', 5)
        self.checkEqual('(let ([x 5] [y 2]) (+ y x))', 7)
        self.checkEqual('(let ([x 5]) (let ([y 2]) (+ y x)))', 7)

    def test_use_after_let(self):
        self.checkRaises('(do (let ([x 5]) x) x)')
        self.checkEqual('(do (define y 2) (let ([x 5]) y))', 2)

    def test_shadow_in_let(self):
        self.checkEqual('(do (define x 2) (let ([x 5]) x))', 5)
        self.checkEqual('(do (define x 2) (let ([x 5]) x) x)', 2)

    def test_undefined_symbol_should_stay_symbol(self):
        self.checkRaises('x')
        self.checkRaises('(do x)')
        self.checkRaises('(do 5 x)')
        self.checkRaises('(define x x)')

    def test_lambda_resolution(self):
        self.checkEqual('((fn [x] x) 5)', 5)
        self.checkEqual('(do (define y 5) ((fn [x] y) 2))', 5)
        self.checkEqual('((fn [x] ((fn [y] x) 2)) 1)', 1)
        self.checkEqual('((fn [x] ((fn [y] y) x)) 11)', 11)
        self.checkEqual('((fn [x y] ((fn [y] y) (+ x y))) 22 33)', 55)
        self.checkEqual('((fn [x y] ((fn [z] y) 0)) 22 33)', 33)

    def test_resolution_in_set(self):
        self.checkRaises('(set [x 5])')
        self.checkEqual('(do (define x 0) (set [x 5]))', 5)
        self.checkEqual('(do (define x 0) (set [x x]))', 0)
        self.checkRaises('(do (define x 0) (set [x y]))')



'''Test wal eval logic'''
import unittest
import random

from wal.core import Wal
from wal.ast_defs import WalEvalError


class BasicOpTest(unittest.TestCase):
    '''Test built-in functions'''

    def setUp(self):
        self.wal = Wal()

    def checkEqual(self, txt, res):
        '''eval first argument and check if result matches second argument '''
        self.assertEqual(self.wal.eval_str(txt), res)

    def test_is_atom(self):
        '''Test atom?'''

        self.checkEqual('(atom? 1)', True)
        self.checkEqual('(atom? "1")', True)
        self.checkEqual("(atom? (+ 1 2))", True)
        self.checkEqual("(atom? '(+ 1 2))", False)
        self.checkEqual("(atom? (array))", False)

    def test_is_symbol(self):
        '''Test symbol?'''

        self.checkEqual("(symbol? 'a)", True)
        self.checkEqual("(symbol? 'TOP.sub.sig)", True)
        self.wal.eval_str("(define x 'a)")
        self.checkEqual("(symbol? x)", True)
        self.checkEqual('(symbol? "symbol")', False)
        self.checkEqual('(symbol? (let ([a 5]) a))', False)
        self.checkEqual('(symbol? 1)', False)
        self.checkEqual("(symbol? '(1 2))", False)
        self.checkEqual("(symbol? (array))", False)

    def test_is_string(self):
        '''Test string?'''

        self.checkEqual('(string? "1")', True)
        self.checkEqual('(string? "1897342hahdasd \\n asa")', True)
        self.checkEqual('(string? (+ "hi" 1))', True)
        self.checkEqual("(string? (symbol->string 'a))", True)

        self.checkEqual('(string? 1)', False)
        self.checkEqual('(let ([a 5]) (string? a))', False)
        self.checkEqual("(string? 'a)", False)
        self.checkEqual("(string? '(1 2))", False)
        self.checkEqual("(string? (array))", False)

    def test_is_int(self):
        '''Test int?'''

        self.checkEqual("(int? 1)", True)
        self.checkEqual("(int? 1142441)", True)
        self.checkEqual("(int? 583547542375894327581)", True)
        self.checkEqual("(int? (+ 1 2))", True)

        self.checkEqual("(int? ''1)", False)
        self.checkEqual("(int? 'a)", False)
        self.checkEqual("(int? '(1 2))", False)
        self.checkEqual('(int? "a")', False)
        self.checkEqual("(int? (array))", False)

    def test_is_list(self):
        '''Test list?'''

        self.checkEqual("(list? '(1 2 3))", True)
        self.checkEqual("(list? (range 10))", True)
        self.checkEqual("(list? '(+ 1 2))", True)

        self.checkEqual("(list? (&& 123))", False)
        self.checkEqual("(list? 'a)", False)
        self.checkEqual('(list? "a")', False)
        self.checkEqual("(list? (array))", False)

    def test_convert_binary(self):
        '''Test convert/bin'''

        self.checkEqual("(convert/bin 0)", "0")
        self.checkEqual("(convert/bin 0 5)", "00000")
        self.checkEqual("(convert/bin 1)", "1")
        self.checkEqual("(convert/bin 1 5)", "00001")
        self.checkEqual("(convert/bin (+ 1 2))", "11")

        for _ in range(50):
            random_int = random.randint(-99999999, 9999999)
            random_width = random.randint(1, 100)
            int_str = self.wal.eval_str(f"(convert/bin {random_int} {random_width})")
            self.assertEqual(random_int, int(int_str, 2))

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(convert/bin "hi")')

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(convert/bin (quote a))')

    def test_string_to_int(self):
        '''Test string->int'''

        self.checkEqual('(string->int "0")', 0)
        self.checkEqual('(string->int "1")', 1)
        self.checkEqual('(string->int "00000")', 0)
        self.checkEqual('(string->int "00001")', 1)
        self.checkEqual('(string->int "101")', 101)
        self.checkEqual('(string->int "-1")', -1)

        for _ in range(50):
            random_int = random.randint(-99999999, 99999999)
            self.checkEqual(f'(string->int "{random_int}")', random_int)

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(string->int (quote a))')

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(string->int 1)')

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(string->int (array))')

    def test_int_to_string(self):
        '''Test int->string'''

        self.checkEqual('(int->string 0)', '0')
        self.checkEqual('(int->string 1)', '1')
        self.checkEqual('(int->string 00000)', '0')
        self.checkEqual('(int->string 00001)', '1')
        self.checkEqual('(int->string 101)', '101')
        self.checkEqual('(int->string -1)', '-1')

        for _ in range(50):
            random_int = random.randint(-99999999, 99999999)
            self.checkEqual(f'(int->string {random_int})', str(random_int))

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(int->string (quote a))')

        with self.assertRaises(WalEvalError):
            self.wal.eval_str('(int->string (array))')



'''Definitions for the wal-awk AST'''

from dataclasses import dataclass
from wal.ast_defs import Symbol as S, Symbol, Operator, WList


@dataclass
class Statement:
    '''Statement class'''
    condition: []
    action: []


@dataclass
class AST:
    '''Abstract syntax tree.
       Contains definitions of functions and statements'''

    def __init__(self, parsed, trace):
        self.functions = {}
        self.statements = []
        self.begin = []
        self.end = []

        for statement in parsed:
            if statement.condition == [S('BEGIN')]:
                self.begin.append(statement.action)
            elif statement.condition == [S('END')]:
                self.end.append(statement.action)
            elif (statement.condition == [True]
                  and isinstance(statement.action, list)
                  and statement.action[0] == Operator.DEFUN):
                self.begin.append(statement.action)
            else:
                self.statements.append(statement)

    def emit(self):
        main_loop = [Operator.WHENEVER, True]
        for stmt in self.statements:
            main_loop.append([S('when'), [Operator.AND, *stmt.condition], stmt.action])

        variables = self.find_variables(self.begin)
        variables = self.find_variables(self.end, vars=variables)
        variables = self.find_variables(main_loop, vars=variables)

        variable_definitions = [[Operator.DEFINE, S(key[0]), key[1]] for key in variables.items()]
        begin_statements = self.begin
        wal = [[Operator.DO, *variable_definitions, *begin_statements]]
        wal.append(main_loop)
        wal += self.end

        symbols = []
        def find_symbols(expr):
            if isinstance(expr, Symbol):
                symbols.append(expr.name)
            elif isinstance(expr, list):
                for sub in expr:
                    find_symbols(sub)

        find_symbols(wal)
        
        return wal, symbols

    def find_variables(self, expr, vars={}):
        if isinstance(expr, (WList, list)):
            if len(expr) > 1 and expr[0] == Operator.SET:
                for binding in expr[1:]:
                    vars[binding[0].name] = 0
                    self.find_variables(binding[1], vars)
            if len(expr) > 1 and expr[0] == Operator.SETA:
                vars[expr[1].name] = [Operator.ARRAY]
                self.find_variables(expr[-1], vars)
            else:
                for sub_expr in expr:
                    self.find_variables(sub_expr, vars)

        return vars



'''Main wal command line tool entry point'''
# pylint: disable=C0103,R0912

import argparse
import os
import sys

from wal.core import Wal
from wal.repl import WalRepl
from wal.util import wal_decode
from wal.reader import read_wal_sexprs, ParseError
from wal.passes import expand, optimize, resolve
from wal.version import __version__ as wal_version
from wal.ast_defs import WalEvalError


class Arguments:  # pylint: disable=too-few-public-methods
    '''Wrapper class for argument parser'''

    def __init__(self):
        parser = argparse.ArgumentParser()
        self.parser = parser

        parser.add_argument('-l', '--load', nargs='*', help='paths to waveforms to load')
        parser.add_argument('-c', nargs='?', help='WAL expression to execute')
        parser.add_argument('program_path', nargs='?', help='path to wal/wo file')
        parser.add_argument('-v', '--version', action='version',
                            version=f'%(prog)s {wal_version}')
        parser.add_argument('--repl-on-failure', action='store_true', help='open REPL when a failure occurs')
        parser.add_argument('ARGS', nargs='*',
                            default=None, help='runtime arguments')


    def parse(self):
        '''Parse program arguments and check minimal requirements'''
        args = self.parser.parse_args()

        return args


def run():
    '''Entry point for the application script'''
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit()


def main():  # pylint: disable=R1710
    '''WAL main function'''
    sys.setrecursionlimit(10000)

    arg_parser = Arguments()
    args = arg_parser.parse()

    wal = Wal()
    wal.eval_context.global_environment.write('ARGS', args.ARGS)

    if args.load is not None:
        for i, path in enumerate(args.load):
            wal.load(path, f't{i}')

    if args.c:
        try:
            wal.eval_str(args.c)
        except Exception as e: # pylint: disable=W0703
            print()
            print('>>>>> Runtime error! <<<<<')
            print(e)
            if args.repl_on_failure:
                WalRepl(wal, intro=WalRepl.dyn_intro).cmdloop()

            return os.EX_SOFTWARE
    else:
        if not args.program_path: # if no arguments where given start REPL
            return WalRepl(wal).cmdloop()

        filename = args.program_path

        if filename[-3:] == '.wo':
            sexprs = wal_decode(filename)
        else:
            with open(filename, 'r', encoding='utf8') as file:
                try:
                    sexprs = read_wal_sexprs(file.read(), filename)
                except ParseError as e:
                    e.show()
                    sys.exit(os.EX_DATAERR)

        try:
            for sexpr in sexprs:
                try:
                    expanded = expand(wal.eval_context, sexpr, parent=wal.eval_context.global_environment)
                    optimized = optimize(expanded)
                    resolved = resolve(optimized, start=wal.eval_context.global_environment.environment)
                except AssertionError as error:
                    wal.print_error(sexpr, error)

                wal.eval(resolved)
        except WalEvalError as error:
            error.print()
            if args.repl_on_failure:
                WalRepl(wal, intro=WalRepl.dyn_intro).cmdloop()

            return os.EX_SOFTWARE

    return os.EX_OK



'''Test wal array eval logic'''
from wal.core import Wal
from wal.ast_defs import WalEvalError
from .test_eval import OpTest

# pylint: disable=C0103
# pylint: disable=W0201

class BasicaArrayTest(OpTest):
    '''Test basic array functions such as list creation'''

    def setUp(self):
        self.w = Wal()
        self.w.context = {
            'a1': {
                '0': 'foo',
                '1': 'bar',
                'foo': 0,
                'bar': 1,
                '0-1': 'bar'
            }
        }

    def test_array_construcion(self):
        '''Test constructing arrays'''
        self.checkEqual('(array (1 2))', {'1': 2})
        self.checkEqual('(array (1 2) (3 4))', {'1': 2, '3': 4})
        self.checkEqual('(array (1 (+ 1 1)))', {'1': 2})
        self.checkEqual('(array ((+ 1 1) 2))', {'2': 2})
        self.checkEqual('(array (1 "a"))', {'1': 'a'})
        self.checkEqual('(array ("a" 1))', {'a': 1})

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(array (1))')
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(array (1 2 3))')

    def test_geta(self):
        '''Test accessing arrays'''
        self.checkEqual('(geta (array (1 2) (3 4)) 1)', 2)
        self.checkEqual('(geta (array (1 2) (3 4)) 3)', 4)

        with self.assertRaises(WalEvalError):
            self.checkEqual('(geta (array (1 2) (3 4)) 4)', 4)

        self.checkEqual("(geta (array (\"a\" 2) ('b 4)) \"a\")", 2)
        self.checkEqual("(geta (array (\"a\" 2) ('b 4)) 'b)", 4)

        self.checkEqual("(geta (array (\"a\" 2) ('a 4)) \"a\")", 4)

        # should fail since geta requires at least 2 arguments
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(geta)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(geta x)')
        # should fail since geta requires first argument to be a symbol
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(geta 5 "a")')

    def test_seta(self):
        '''Test modifying arrays'''
        self.checkEqual("(seta (array) 'a 2)", {'a': 2})
        self.checkEqual("(seta (array) \"a\" 2)", {'a': 2})
        self.checkEqual("(seta (array) 12 2)", {'12': 2})
        self.checkEqual("(seta (array ['a 5]) 'a 2)", {'a': 2})

        # should fail since first argument must be a symbol or array
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(seta 0 0)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(seta "a" 0)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(seta '(1 2) 0)")
        # should fail since keys must be either string or int
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(seta (array) '(1 2) 2)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(seta (array) (array) 2)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(seta (array) '+ 2)")



'''Test wal eval logic'''
import unittest
import math
from unittest.mock import patch
from io import StringIO

from wal.core import Wal
from wal.reader import read_wal_sexpr
from wal.ast_defs import Symbol as S
from wal.ast_defs import Operator
from wal.ast_defs import WalEvalError


class OpTest(unittest.TestCase):
    '''Test built-in functions'''

    def setUp(self):
        self.w = Wal()
        self.w.eval_context.environment.define('x', 5)
        self.w.eval_context.environment.define('y', 10)
        self.w.eval_context.environment.define('z', 0)

    @property
    def x(self):
        return self.w.eval_context.environment.read('x')

    @property
    def y(self):
        return self.w.eval_context.environment.read('y')

    @property
    def z(self):
        return self.w.eval_context.environment.read('z')

    def checkEqual(self, txt, res):
        '''eval first argument and check if result matches second argument '''
        sexpr = read_wal_sexpr(txt)
        self.assertEqual(self.w.eval(sexpr), res)


class BasicOpTest(OpTest):
    '''Test built-in functions'''

    def test_add(self):
        '''Test add evaluation'''

        # test int addition
        self.checkEqual('(+ 1 2)', 3)
        self.checkEqual('(+ x y)', self.x + self.y)
        self.checkEqual('(+ x 1)', self.x + 1)
        self.checkEqual('(+ x -1)', self.x + -1)
        # test mixed int string
        self.checkEqual('(+ x "hi")', f'{self.x}hi')
        # test string concatenation
        self.checkEqual('(+ "hi" "ho")', 'hiho')
        # test nested additions
        self.checkEqual('(+ (+ 1 1) (+ 2 2))', 6)
        # test single argument
        self.checkEqual('(+ 1)', 1)
        # test no argument
        self.checkEqual('(+)', 0)

        # test list append
        self.checkEqual("(+ '(1 2) '(3 4))", [1, 2, 3, 4])
        self.checkEqual("(+ '(1 2) 3)", [1, 2, 3])

    def test_sub(self):
        '''Test sub evaluation'''

        # test int sub
        self.checkEqual('(- 5)', -5)
        self.checkEqual('(- 2 1)', 1)
        self.checkEqual('(- 1 2)', -1)
        self.checkEqual('(- y x)', self.y - self.x)
        # test mixed int string
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(- x 1 "test")')

        # test nested additions
        self.checkEqual('(- 10 (- 4 2))', 8)
        self.checkEqual('(- (- 10 10) (- 4 2))', -2)

    def test_mul(self):
        '''Test mul evaluation'''
        self.checkEqual('(* 2 2)', 4)
        self.checkEqual('(* x y)', self.x * self.y)
        self.checkEqual('(* x y z)', self.x * self.y * self.z)

    def test_div(self):
        '''Test mul evaluation'''
        self.checkEqual('(/ 22 2)', 11)
        self.checkEqual('(/ y x)', self.y / self.x)

        # we dont raise an exception but just print a warning
        # since this makes wal code significantly cleaner in some cases
        # with self.assertRaises(ZeroDivisionError):
        #     self.w.eval('(/ 5 0)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(/ 1)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(/ 1 "a")')

    def test_exp(self):
        '''Test exp evaluation'''
        self.checkEqual('(** 2 10)', 1024)
        self.checkEqual('(** x y)', math.pow(self.x, self.y))

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(** 1)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(** 1 "a")')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(** 1 2 3)')

    def test_eq(self):
        '''Test eq operator'''
        self.checkEqual('(= 3 3)', 1)
        self.checkEqual('(= 3 3 3)', 1)
        self.checkEqual('(= "a" "a")', 1)
        self.checkEqual('(= 3 3 "a")', 0)
        self.checkEqual('(= x x)', 1)
        self.checkEqual('(= x y)', 0)
        self.checkEqual('(= (+ 2 2) (* 2 2))', 1)

    def test_neq(self):
        '''Test neq operator'''
        self.checkEqual('(!= 3 3)', 0)
        self.checkEqual('(!= 3 3 3)', 0)
        self.checkEqual('(!= "a" "a")', 0)
        self.checkEqual('(!= 3 3 "a")', 1)
        self.checkEqual('(!= x x)', 0)
        self.checkEqual('(!= x y)', 1)
        self.checkEqual('(!= (+ 2 2) (* 2 2))', 0)


    def test_larger(self):
        '''Test larger operator'''
        self.checkEqual('(> 1 0)', 1)
        self.checkEqual('(> 2 1)', 1)
        self.checkEqual('(> 1 -1)', 1)
        self.checkEqual('(> 0 0)', 0)
        self.checkEqual('(> -1 0)', 0)
        self.checkEqual('(> -2 -1)', 0)
        self.checkEqual('(> (+ 5 5) 5)', 1)
        self.checkEqual('(> x y)', 0)

        # compare only on two elements
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(> 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(> 3 4 5)")

        # compare only on ints
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(> '(1 2) 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(> '(1 2) '(3))")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(> 3 '(3))")


    def test_smaller(self):
        '''Test smaller operator'''
        self.checkEqual('(< 1 0)', 0)
        self.checkEqual('(< 2 1)', 0)
        self.checkEqual('(< 1 -1)', 0)
        self.checkEqual('(< 0 0)', 0)
        self.checkEqual('(< -1 0)', 1)
        self.checkEqual('(< -2 -1)', 1)
        self.checkEqual('(< (+ 5 5) 5)', 0)
        self.checkEqual('(< x y)', 1)

        # compare only on two elements
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(< 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(< 3 4 5)")

        # compare only on ints
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(< '(1 2) 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(< '(1 2) '(3))")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(< 3 '(3))")

    def test_larger_equal(self):
        '''Test larger equals operator'''
        self.checkEqual('(>= 1 0)', 1)
        self.checkEqual('(>= 2 1)', 1)
        self.checkEqual('(>= 1 -1)', 1)
        self.checkEqual('(>= 0 0)', 1)
        self.checkEqual('(>= -1 0)', 0)
        self.checkEqual('(>= -2 -1)', 0)
        self.checkEqual('(>= (+ 5 5) 5)', 1)
        self.checkEqual('(>= x y)', 0)

        # compare only on two elements
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(>= 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(>= 3 4 5)")

        # compare only on ints
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(>= '(1 2) 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(>= '(1 2) '(3))")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(>= 3 '(3))")

    def test_smaller_equal(self):
        '''Test smaller equals operator'''
        self.checkEqual('(<= 1 0)', 0)
        self.checkEqual('(<= 2 1)', 0)
        self.checkEqual('(<= 1 -1)', 0)
        self.checkEqual('(<= 0 0)', 1)
        self.checkEqual('(<= -1 0)', 1)
        self.checkEqual('(<= -2 -1)', 1)
        self.checkEqual('(<= (+ 5 5) 5)', 0)
        self.checkEqual('(<= x y)', 1)

        # compare only on two elements
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(<= 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(<= 3 4 5)")

        # compare only on ints
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(<= '(1 2) 3)")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(<= '(1 2) '(3))")
        with self.assertRaises(WalEvalError):
            self.w.eval_str("(<= 3 '(3))")

    def test_and(self):
        '''Test logical and operator'''
        self.checkEqual('(&& 1 1)', True)
        self.checkEqual('(&& 1 1 2)', True)
        self.checkEqual('(&& 1 "a")', True)
        self.checkEqual("(&& 1 '(1 2 3))", True)
        self.checkEqual('(&& 1 0)', False)
        self.checkEqual('(&& 0 0)', False)

    def test_or(self):
        '''Test logical and operator'''
        self.checkEqual('(|| 1 1)', True)
        self.checkEqual('(|| 1 1 2)', True)
        self.checkEqual('(|| 1 "a")', True)
        self.checkEqual("(|| 1 '(1 2 3))", True)
        self.checkEqual('(|| 1 0)', True)
        self.checkEqual('(|| 0 0)', False)

    def test_not(self):
        '''Test not operator'''
        self.checkEqual('(! 1)', 0)
        self.checkEqual('(! 0)', 1)

        with self.assertRaises(WalEvalError):
            self.w.eval_str("(!)")

    def test_set(self):
        '''Test variable set function'''
        self.checkEqual('x', self.x)
        self.w.eval_str('(set (x 6))')
        self.checkEqual('x', 6)

        self.w.eval_str('(set (x (+ 1 2)))')
        self.checkEqual('x', 3)

        self.w.eval_str('(set (x (+ x x)))')
        self.checkEqual('x', 6)

        # check if set returns correct value
        self.checkEqual('(set (x (+ x x)))', 12)

        self.w.eval_str('(do (define a 0) (set (a 11)))')
        self.checkEqual('a', 11)
        self.w.eval_str('(do (define b 1) (set (a 11) (b 22)))')
        self.checkEqual('b', 22)

        self.checkEqual('(let ([b 0]) (set (a 11) (b 22)))', 22)

        with self.assertRaises(WalEvalError):
            # should fail because only symbols can be assigned to
            self.w.eval_str('(set (5 5))')

            # should fail because xx is not defined
            self.w.eval_str('(set (xx 5))')


class EvalPrintTest(OpTest):
    '''Test functions that produce output'''

    def test_print(self):
        '''Test simple print function'''
        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(print "hello, world!")')
            self.assertEqual(fake_out.getvalue(), 'hello, world!\n')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(print "a" "bc")')
            self.assertEqual(fake_out.getvalue(), 'abc\n')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(print "a" "bc" 123)')
            self.assertEqual(fake_out.getvalue(), 'abc123\n')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(print)')
            self.assertEqual(fake_out.getvalue(), '\n')

    def test_printf(self):
        '''Test printf formated output'''
        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "hello, world!")')
            self.assertEqual(fake_out.getvalue(), 'hello, world!')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "%s" "hello, world!")')
            self.assertEqual(fake_out.getvalue(), 'hello, world!')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "%d" 15)')
            self.assertEqual(fake_out.getvalue(), '15')
            self.w.eval_str('(printf "%04d" 15)')
            self.assertEqual(fake_out.getvalue(), '150015')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "%x" 15)')
            self.assertEqual(fake_out.getvalue(), 'f')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "%x\\n" 15)')
            self.assertEqual(fake_out.getvalue(), 'f\n')

        with patch('sys.stdout', new=StringIO()) as fake_out:
            self.w.eval_str('(printf "")')
            self.assertEqual(fake_out.getvalue(), '')

        with self.assertRaises(ValueError):
            self.w.eval_str('(printf 5 5)')


class EvalControlFlowTest(OpTest):
    '''Test control flow functions'''

    def test_if(self):
        '''Test if statements'''
        self.checkEqual('(if 1 2 3)', 2)
        self.checkEqual('(if 0 2 3)', 3)
        self.checkEqual('(if (> x y) 2 3)', 3)
        self.checkEqual('(if (! (> x y)) 2 3)', 2)
        self.checkEqual('(if 1 (+ 1 1) 3)', 2)
        self.checkEqual('(if 0 2 (+ 1 2))', 3)

        self.checkEqual('(if 1 2)', 2)
        self.checkEqual('(if 0 2)', None)

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(if)')
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(if 1)')

    def test_cond(self):
        '''Test conditional branches'''
        self.checkEqual('(cond (1 2) (2 3))', 2)
        self.checkEqual('(cond (0 2) (2 3))', 3)
        self.checkEqual('(cond (0 2) (0 3))', None)
        self.checkEqual('(cond ((= 1 1) 2) (2 3))', 2)
        self.checkEqual('(cond ((= 1 1) (+ 1 1)) (2 3))', 2)

    def test_when(self):
        '''Test when construct'''
        self.checkEqual('(when 1 2)', 2)
        self.checkEqual('(when 0 2)', None)
        self.checkEqual('(when (= 1 1) 2)', 2)
        self.checkEqual('(when (= 1 0) 2)', None)
        self.checkEqual('(when (= 1 1) (+ 1 2))', 3)

    def test_unless(self):
        '''Test when construct'''
        self.checkEqual('(unless 1 2)', None)
        self.checkEqual('(unless 0 2)', 2)
        self.checkEqual('(unless (= 1 1) 2)', None)
        self.checkEqual('(unless (= 1 0) 2)', 2)
        self.checkEqual('(unless (= 1 1) (+ 1 2))', None)

    def test_case(self):
        '''Test case construct'''
        case1 = '(case a (1 "a") (2 "b") (3 "c"))'
        self.w.eval_str('(define a 0)')
        self.w.eval_str('(set (a 1))')
        self.checkEqual(case1, "a")
        self.w.eval_str('(set (a 2))')
        self.checkEqual(case1, "b")
        self.w.eval_str('(set (a 3))')
        self.checkEqual(case1, "c")

        self.w.eval_str('(set (a 4))')
        self.checkEqual(case1, None)

        # match complex data
        case2 = '(case a ("x" "a") ("y" "b") ("z" "c"))'
        self.w.eval_str('(set (a "x"))')
        self.checkEqual(case2, "a")

        case3 = '(case a ((1 1) "a") ((1 2) "b"))'
        self.w.eval_str("(set (a '(1 1)))")
        self.checkEqual(case3, "a")

        with self.assertRaises(ValueError):
            self.w.eval_str('(case 1 (1 a) (2 b) (1 c))')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(case)')

    def test_do(self):
        '''Test sequential program execution'''
        self.checkEqual('(do 1 2)', 2)
        self.checkEqual('(do (+ 1 2) (+ 2 2))', 4)

    def test_while(self):
        '''Test while loop'''
        self.w.eval_str('(set (x 0))')
        self.checkEqual('(while (< x 5) (set (x (+ x 1))))', 5)

        # should fail since while expects 2 args
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(while (< x 5))')

    def test_alias(self):
        '''Test symbol renaming using aliases'''
        self.w.eval_str('(alias abc \'x)')
        self.checkEqual('abc', self.x)

        self.w.eval_str('(alias abc "y")')
        self.checkEqual('abc', self.y)

        # alias expects two arguments
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(alias)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(alias a)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(alias a 1 b)')

        # first argument must be a symbol
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(alias 0 "a")')

    def test_unalias(self):
        '''Test deleting aliases using unalias'''
        self.w.eval_str('(alias abc \'x)')
        self.checkEqual('abc', self.x)

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(unalias abc)')
            self.w.eval_str('abc')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(unalias)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(unalias 1)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(unalias x y)')

    def test_quote(self):
        '''Test quoting'''
        self.checkEqual('(quote x)', S('x'))
        self.checkEqual("'x", S('x'))
        self.checkEqual("'(+ 1 2)", [Operator.ADD, 1, 2])

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(quote)')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(quote a b)')


class EvalFunctionTest(OpTest):
    '''Test control flow functions'''

    def test_lambda_args(self):
        '''Lambdas should raise errors for wrong arguments'''
        with self.assertRaises(WalEvalError):
            self.w.eval_str('(fn (x))')

        with self.assertRaises(WalEvalError):
            self.w.eval_str('(fn (x 1) x)')

    def test_lambda_apply(self):
        '''Lambdas should perform correct action if applied'''

        # Apply from inline definition
        self.checkEqual('((fn (x) (+ x 1)) 1)', 2)

        # correct number of args must be supplied
        # with self.assertRaises(WalEvalError):
        #     self.w.eval_str('(lambda (x) (+ x 1) 1 2)')

        # Apply named lambda
        self.w.eval_str('(define foo (fn (y) (* y 2)))')
        self.checkEqual('(foo 5)', 10)


class MultipleTracesTest(unittest.TestCase):
    '''Test multiple traces'''

    def setUp(self):
        self.w = Wal()
        self.w.load('tests/traces/counter.fst', tid='a')
        self.w.load('tests/traces/counter.vcd', tid='b')

    def checkEqual(self, txt, res):
        '''eval first argument and check if result matches second argument '''
        sexpr = read_wal_sexpr(txt)
        self.assertEqual(self.w.eval(sexpr), res)

    def test_index(self):
        self.w.eval_str('(define x 0)')
        self.w.eval_str('(whenever (!= a^INDEX b^INDEX) (inc x))')
        self.checkEqual('x', 0)



'''WAL Interpreter Passes '''
from math import prod

from wal.ast_defs import Operator, Symbol, Macro, Environment, WList
from wal.util import wal_str


def recursive_set_line_info(expr, line_info):
  if isinstance(expr, (WList, list)):
    return WList([recursive_set_line_info(sub, line_info) for sub in expr], line_info=line_info)
  elif isinstance(expr, Symbol):
    expr.line_info = line_info
    return expr
  else:
    return expr


def expand(seval, exprs, parent=None):
    '''Macroexpansion Pass'''
    if isinstance(exprs, (WList, list)):
        if len(exprs) > 0 and exprs[0] in [Operator.QUOTE, Operator.QUASIQUOTE]:
            return exprs
        if len(exprs) > 0 and isinstance(exprs[0], Symbol) and seval.environment.is_defined(exprs[0].name):
            expr = seval.environment.read(exprs[0].name)
            if isinstance(expr, Macro):
                macro_env = Environment(parent=parent)
                vals = exprs[1:]

                if isinstance(expr.args, Symbol):
                    macro_env.define(expr.args.name, vals)
                elif isinstance(expr.args, WList):
                    assert len(expr.args) == len(vals), f'{expr.name}: number of passed arguments does not match expected number'
                    for arg, val in zip(expr.args, vals):
                        macro_env.define(arg.name, val)
                else:
                    assert False, f'cannot evaluate {wal_str(expr)}'

                save_env = seval.environment
                seval.environment = macro_env
                expanded = seval.eval(expr.expression)
                expanded.line_info = exprs[0].line_info
                expanded = expand(seval, expanded, parent)

                seval.environment = save_env

                if isinstance(expanded, WList):
                    exprs.clear()
                    for expr in expanded:
                        exprs.append(expr)
                else:
                    return expanded


        line_info = exprs.line_info if isinstance(exprs, WList) else None
        exprs = WList(list(map(lambda expr: expand(seval, expr, parent=parent), exprs)), line_info=line_info)
    
    return exprs


def optimize(expr):
    '''Optimization Pass'''
    try:
        if isinstance(expr, (WList, list)):
            op = expr[0]
            if op in [Operator.QUOTE, Operator.QUASIQUOTE]:
                pass
            elif op == Operator.IF:
                expr = WList(list(map(optimize, expr)), line_info=expr.line_info)
                if (expr[1] is True) or isinstance(expr[1], (int, float, str)):
                    if expr[1]:
                        # (if #t then else) => then
                        expr = expr[2]
                    else:
                        # (if #f/0 then else) => else
                        expr = expr[3]
            # (do expr) => expr
            elif op == Operator.DO:
                expr = WList(list(map(optimize, expr)), line_info=expr.line_info)
                if len(expr) == 2:
                    expr = optimize(expr[1])

            elif op == Operator.ADD:
                expr = WList(list(map(optimize, expr)), line_info=expr.line_info)
                if all(isinstance(arg, (int, float)) for arg in expr[1:]):
                    expr = sum(expr[1:])
                elif all(isinstance(arg, str) for arg in expr[1:]):
                    expr = ''.join(expr[1:])

            elif op == Operator.MUL:
                expr = WList(list(map(optimize, expr)), line_info=expr.line_info)
                if any(arg == 0 for arg in expr[1:]):
                    expr = 0
                if all(isinstance(arg, (int, float)) for arg in expr[1:]):
                    expr = prod(expr[1:])

            elif op == Operator.AND:
                # (&& x) => x
                if len(expr) == 2:
                    expr = expr[1]
                elif any(not arg for arg in expr[1:]):
                    expr = False
                elif all(isinstance(arg, (int, float, str)) and arg != 0 for arg in expr[1:]):
                    expr = True

            elif op == Operator.OR:
                # (|| x) => x
                if len(expr) == 2:
                    expr = expr[1]
                elif any(isinstance(arg, (int, float, str)) and arg != 0 for arg in expr[1:]):
                    expr = True
                elif all(not arg for arg in expr[1:]):
                    expr = False
            else:
                expr = WList(list(map(optimize, expr)), line_info=expr.line_info)

    finally:
        return expr


def resolve(expr, start={}):
    '''Variable environment resolution pass '''
    scopes=[None, dict(start)]
    def resolve_vars(expr):
        if isinstance(expr, WList) and len(expr) > 0:
            op = expr[0]
            if op == Operator.DEFINE:
                id = expr[1]
                assert id.name not in scopes[-1], f'symbol {id} already defined'
                body = resolve_vars(expr[2])
                scopes[-1][expr[1].name] = True
                return WList([Operator.DEFINE, id, body], line_info=expr.line_info)
            elif op == Operator.LET:
                env = {}
                scopes.append(env)
                for binding in expr[1]:
                    env[binding[0].name] = True

                body = [resolve_vars(sub) for sub in expr[2:]]
                scopes.pop()
                return WList([Operator.LET, expr[1], *body], line_info=expr.line_info)
            elif op == Operator.FN:
                args = expr[1]
                env = {}
                scopes.append(env)
                if isinstance(args, (WList, list)):
                    for arg in expr[1]:
                        assert isinstance(arg, Symbol), 'fn: parameters must be symbols'
                        env[arg.name] = True
                elif isinstance(args, Symbol):
                    env[args.name] = True
                else:
                    assert False, 'fn: first argument must be a list or a symbol'

                body = [resolve_vars(sub) for sub in expr[2:]]
                scopes.pop()
                return WList([Operator.FN, expr[1], *body])
            elif op == Operator.DEFMACRO:
                scopes[-1][expr[1].name] = True
                return expr
            elif op in [Operator.QUOTE, Operator.QUASIQUOTE, Operator.ALIAS]:
                return expr
            else:
                return WList([resolve_vars(sub) for sub in expr], line_info=expr.line_info)
        elif isinstance(expr, Symbol):
            id = expr.name
            steps = 0
            while scopes[-steps-1] is not None and id not in scopes[-steps-1]:
                steps += 1

            if scopes[-steps-1]:
                return Symbol(expr.name, steps)
            else:
                # if the symbol is not defined it still can be a signal
                # in this case just use regular slow resolution
                return expr

        # all other expressions can just be returned as they are
        return expr

    return(resolve_vars(expr))




'''Implementation of basic read-eval-print-loop'''
import cmd
try:
    import readline
except (ModuleNotFoundError, ImportError):
    readline = None

import os

from wal.util import wal_str
from wal.reader import read_wal_sexpr, ParseError
from wal.ast_defs import Operator, Symbol, WList
from wal.trace.trace import Trace
from wal.version import __version__

completer_delims = ' \t\n`~@#()[{]}\'"'

histfile = os.path.expanduser('~/.wal/.wal_history')
histfile_size = 1000

class WalRepl(cmd.Cmd):
    '''WalRepl class implements basic read-eval-print-loop'''


    std_intro = f'''WAL {__version__}
Exit to OS or terminate running evaluations with CTRL-C'''
    dyn_intro = f'''WAL {__version__}
Started interactive WAL REPL.
Exit to calling script or terminate running evaluations with CTRL-C'''

    terminator = ['(', ')']
    complete_list = [op.value for op in Operator]
    file = None

    def __init__(self, wal, intro=std_intro):
        super().__init__()
        self.wal = wal
        self.intro = intro

    @property
    def prompt(self):
        '''Generate prompt symbol'''
        ids = [f'{k}({self.wal.traces.traces[k].index})' for k in self.wal.traces.traces.keys()]
        traces = ', '.join(ids)
        if self.wal.traces.traces:
            traces = traces + ' '
        return f'{traces}>-> '


    def onecmd(self, line):
        try:
            evaluated = self.wal.eval(line)

            if evaluated is not None:
                print(wal_str(evaluated))

            if readline:
                readline.set_completer_delims(completer_delims)
                readline.set_history_length(histfile_size)
                readline.write_history_file(histfile)

        except KeyboardInterrupt:
            print('Keyboard Interrupt')
        except Exception as e:
            print(e)
            print(wal_str(line))


    def precmd(self, line):
        try:
            sexpr = read_wal_sexpr(line)
            # intercept defuns to include them in completion
            if isinstance(sexpr, WList):
                if sexpr[0] == Symbol('defun'):
                    self.complete_list.append(sexpr[1].name)

            return sexpr
        except ParseError as e:
            e.show()
        except Exception as e:
            print(e)
        return None

    def complete(self, text, state):
        return self.completenames(text)[state]

    def completenames(self, text, *ignored):
        tmp = self.complete_list + self.wal.traces.signals + list(self.wal.eval_context.global_environment.environment.keys())

        if len(self.wal.traces.traces) == 1:
            tmp += Trace.SPECIAL_SIGNALS

        candidates = [c for c in tmp if c.startswith(text)]
        candidates.append(None)
        return candidates

    def preloop(self):
        if not os.path.exists(histfile):
            os.makedirs(os.path.expanduser('~/.wal'), exist_ok=True)

        if readline:
            readline.set_completer_delims(completer_delims)
            if os.path.exists(histfile):
                readline.read_history_file(histfile)



'''Implementations for special hardware related functions'''
from wal.ast_defs import Operator, Symbol, WList

def op_find(seval, args):
    '''Find
       Returns a list of all indices at which the condition in argument 1
       is true. Steps each trace individually.
    '''
    assert len(args) == 1, 'find: expects exactly one argument (find condition)'

    found = []
    for trace in seval.traces.traces.values():
        start_index = trace.index # store current index
        ended = False
        while not ended:
            if seval.eval(args[0]):
                found.append(trace.index)
            ended = trace.step()

        trace.index = start_index # reset trace index

    found = list(set(found))
    found.sort()
    return found


def op_find_g(seval, args):
    '''Find Global
       Returns a list of all indices at which the condition in argument 1
       is true. Steps all traces at the same time to allow conditions defined
       over all traces.
    '''
    assert len(args) == 1, 'find: expects exactly one argument (find condition)'

    prev_indices = seval.traces.indices()
    found = []
    ended = []
    while not ended:
        if seval.eval(args[0]):
            indices = seval.traces.indices()
            found.append(indices if len(indices) > 1 else list(indices.values())[0])

        ended = seval.traces.step()

    for trace in seval.traces.traces.values():
        trace.index = prev_indices[trace.tid]

    return found


def op_whenever(seval, args):
    '''Evaluates body at each index at which condition evaluate to true '''
    assert len(args) >= 2, 'whenever: expects exactly two arguments (whenever condition body)'

    prev_indices = seval.traces.indices()

    res = None
    ended = []
    while not ended:
        if seval.eval(args[0]):
            res = seval.eval_args(args[1:])[-1]

        ended = seval.traces.step()

    for trace in seval.traces.traces.values():
        trace.index = prev_indices[trace.tid]

    return res


def op_fold_signal(seval, args):
    '''Performs a fold operation on the values of signal from index INDEX
    until the stop condition evaluates to true. '''
    assert len(args) == 4, 'fold/signal: expects 3 arguments (fold f acc stop signal)'
    func = seval.eval(args[0])
    assert isinstance(func, list) and (func[0] == Operator.FN), 'fold/signal: not a valid function'
    acc = seval.eval(args[1])
    stop = args[2]
    signal = seval.eval(args[3])
    assert isinstance(signal, Symbol), 'fold/signal: last argument must be a signal'
    assert seval.traces.contains(signal.name), f'fold/signal: signal "{signal.name}" not found'

    # store indices at start
    prev_indices = seval.traces.indices()

    stopped = False
    while not seval.eval(stop) and not stopped:
        if isinstance(func, Operator):
            acc = seval.eval(WList([func, WList([Operator.QUOTE, acc]), WList([Operator.QUOTE, seval.eval(signal)])]))
        else:
            assert func[0] == Operator.FN, 'fold/signal: first argument must be a function'
            acc = seval.eval_lambda(func, [[Operator.QUOTE, acc], [Operator.QUOTE, seval.eval(signal)]])

        stopped = seval.traces.step() != []

    # restore indices to start values
    for trace in seval.traces.traces.values():
        trace.index = prev_indices[trace.tid]

    return acc


def op_signal_width(seval, args):
    '''Returns the width of a signal'''
    assert len(args) == 1, 'signal-width: expects exactly one argument (signal-width signal:str?)'
    arg = seval.eval(args[0])
    assert isinstance(arg, (str, Symbol)), 'signal-width: expects exactly one argument (signal-width signal:str?)'
    name = arg if isinstance(arg, str) else arg.name
    assert seval.traces.contains(name), f'signal-width: no signal "{name}"'
    return seval.traces.signal_width(name)


def op_sample_at(seval, args):
    '''Sets the timestamps of trace t to list xs'''
    assert len(args) == 1 or len(args) == 2, 'sample-at: expects one or two arguments (sample-at timestamps:list? trace:symbol?)'
    timestamps = seval.eval(args[0])
    assert isinstance(timestamps, (WList, list)), 'sample-at: first argument must be a list of integers'
    assert all(map(lambda x: isinstance(x, int), timestamps)), 'sample-at: second argument must be a list of integers'

    if len(args) == 2:
        tid = args[1]
        assert isinstance(tid, Symbol), 'sample-at: second argument must be a symbol'
        seval.traces.traces[tid.name].set_sampling_points(timestamps)
    else:
        for trace in seval.traces.traces.values():
            trace.set_sampling_points(timestamps)

def op_trim_trace(seval, args):
    '''Trims the trace to end at some point'''
    assert len(args) == 2, 'trim-trace: expects exactly one argument (trim-trace t:symbol?|str? e:int?)'
    tid = seval.eval(args[0])
    new_max_index = seval.eval(args[1])
    assert isinstance(tid, (str, Symbol)), 'trim-trace: first argument must evaluate to symbol or string'
    assert isinstance(new_max_index, int), 'trim-trace: second argument must evaluate to int'
    tid = tid.name if isinstance(tid, str) else tid.name
    return seval.traces.traces[tid].set_max_index(new_max_index)


special_operators = {
    Operator.FIND.value: op_find,
    Operator.FIND_G.value: op_find_g,
    Operator.WHENEVER.value: op_whenever,
    Operator.FOLD_SIGNAL.value: op_fold_signal,
    Operator.SIGNAL_WIDTH.value: op_signal_width,
    Operator.SAMPLE_AT.value: op_sample_at,
    Operator.TRIM_TRACE.value: op_trim_trace,
}



'''Implementations of math functions'''
# pylint: disable=C0116
import operator

from functools import reduce
from wal.ast_defs import Operator


def op_bor(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int)), evaluated))
    if len(evaluated) == 1:
        res = evaluated[0]
    else:
        res = reduce(operator.__or__, evaluated)
    return res

def op_band(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int)), evaluated))
    if len(evaluated) == 1:
        res = evaluated[0]
    else:
        res = reduce(operator.__and__, evaluated)
    return res

def op_bxor(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int)), evaluated))
    if len(evaluated) == 1:
        res = evaluated[0]
    else:
        res = reduce(operator.__xor__, evaluated)
    return res

bitwise_operators = {
    Operator.BOR.value: op_bor,
    Operator.BAND.value: op_band,
    Operator.BXOR.value: op_bxor,
}



''' Utility functions for WAL'''
# pylint: disable=R0912

import pickle
from wal.ast_defs import Symbol, Operator, Closure, Macro, Unquote, UnquoteSplice, WList


def wal_str(sexpr):
    '''Returns a string representation of a WAL expression'''
    if isinstance(sexpr, (list, WList)):
        if len(sexpr) == 2 and sexpr[0] == Operator.QUOTE:
            txt = f"'{wal_str(sexpr[1])}"
        elif len(sexpr) == 2 and sexpr[0] == Operator.QUASIQUOTE:
            txt = f"`{wal_str(sexpr[1])}"
        elif len(sexpr) == 2 and sexpr[0] == Operator.UNQUOTE:
            txt = f",{wal_str(sexpr[1])}"
        elif len(sexpr) == 3 and sexpr[0] == Symbol('reval'):
            txt = f'{wal_str(sexpr[1])}@{sexpr[2]}'
        elif len(sexpr) > 0 and sexpr[0] == Operator.ARRAY:
            txt = '{' + ' '.join(map(wal_str, sexpr)) + '}'
        else:
            txt = '(' + ' '.join(map(wal_str, sexpr)) + ')'
    elif isinstance(sexpr, Symbol):
        txt = sexpr.name
    elif isinstance(sexpr, Macro):
        txt = f'Macro: {sexpr.name}\nArgs: {wal_str(sexpr.args)}\n' + wal_str(sexpr.expression)
    elif isinstance(sexpr, Closure):
        txt = f'Function: {sexpr.name}\nArgs: {wal_str(sexpr.args)}\n' + wal_str(sexpr.expression)
    elif isinstance(sexpr, Unquote):
        txt = f',{wal_str(sexpr.content)}'
    elif isinstance(sexpr, UnquoteSplice):
        txt = f',@{wal_str(sexpr.content)}'
    elif isinstance(sexpr, Operator):
        txt = sexpr.value
    elif isinstance(sexpr, str):
        sexpr = sexpr.replace("\"", "\\\"")
        txt = f'"{sexpr}"'
    elif isinstance(sexpr, bool):
        txt = 'true' if sexpr else 'false'
    elif isinstance(sexpr, dict):
        content = []
        for key, value in sexpr.items():
            content.append('("' + key + '" ' + wal_str(value) + ')')

        txt = '{' + ' '.join(content) + '}'
    else:
        txt = str(sexpr)

    return txt


def wal_decode(filename):
    '''Decodes a compiled WAL file and returns its WAL expressions'''
    with open(filename, 'br') as fin:
        return pickle.load(fin)


class Colors:
    RED = '\033[0;31m'
    GREEN = '\033[0;32m'
    YELLOW = '\033[0;33m'
    BLUE = "\033[0;34m"
    PURPLE = "\033[0;35m"
    HEADER = '\033[95m'
    END = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    def red(txt):
        return f'{Colors.RED}{txt}{Colors.END}'

    def green(txt):
        return f'{Colors.GREEN}{txt}{Colors.END}'

    def yellow(txt):
        return f'{Colors.YELLOW}{txt}{Colors.END}'

    def blue(txt):
        return f'{Colors.BLUE}{txt}{Colors.END}'

    def purple(txt):
        return f'{Colors.PURPLE}{txt}{Colors.END}'

    def header(txt):
        return f'{Colors.HEADER}{txt}{Colors.END}'

    def bold(txt):
        return f'{Colors.BOLD}{txt}{Colors.END}'        
    
    def underline(txt):
        return f'{Colors.UNDERLINE}{txt}{Colors.END}'        



'''Implementations of math functions'''
# pylint: disable=C0116
import math
import operator

from functools import reduce
from wal.ast_defs import Operator, WList


def op_add(seval, args):
    evaluated = seval.eval_args(args)
    if any(map(lambda x: isinstance(x, (WList, list)), evaluated)):
        res = []
        for item in evaluated:
            if isinstance(item, (WList, list)):
                res += item
            else:
                res.append(item)
    elif any(map(lambda x: isinstance(x, str), evaluated)):
        res = ''.join(map(str, evaluated))
    else:
        res = sum(evaluated)
    return res


def op_sub(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int, float)), evaluated))
    if len(evaluated) == 1:
        res = -evaluated[0]
    else:
        res = reduce(operator.__sub__, evaluated)
    return res


def op_mul(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int, float)), evaluated))
    assert len(evaluated) > 1
    return reduce(operator.__mul__, evaluated)


def op_div(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int, float)), evaluated))
    assert len(evaluated) == 2
    assert evaluated[1] != 0, 'div: division by zero'
    return evaluated[0] / evaluated[1]


def op_exp(seval, args):
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int, float)), evaluated))
    assert len(evaluated) == 2
    return int(math.pow(evaluated[0], evaluated[1]))


def op_floor(seval, args):
    assert len(args) == 1, 'floor: expects exactly one argument (floor x:num?)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (int, float)), 'floor: expects exactly one argument (floor x:num?)'
    return math.floor(evaluated)


def op_ceil(seval, args):
    assert len(args) == 1, 'floor: expects exactly one argument (ceil x:num?)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (int, float)), 'ceil: expects exactly one argument (ceil x:num?)'
    return math.ceil(evaluated)


def op_round(seval, args):
    assert len(args) == 1, 'round: expects exactly one argument (round x:num?)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (int, float)), 'round: expects exactly one argument (round x:num?)'
    return round(evaluated)


def op_mod(seval, args):
    assert len(args) == 2, 'mod: expects exactly one argument (mod x:num?)'
    evaluated = seval.eval_args(args)
    assert all(map(lambda x: isinstance(x, (int, float)), evaluated)), 'mod: expects exactly one argument (mod x:num? y:num?)'
    return evaluated[0] % evaluated[1]


math_operators = {
    Operator.ADD.value: op_add,
    Operator.SUB.value: op_sub,
    Operator.MUL.value: op_mul,
    Operator.DIV.value: op_div,
    Operator.EXP.value: op_exp,
    Operator.FLOOR.value: op_floor,
    Operator.CEIL.value: op_ceil,
    Operator.ROUND.value: op_round,
    Operator.MOD.value: op_mod,
}



'''Implementations for all basic core functions'''
# pylint: disable=C0116,C0103
import re
import os
import sys
import importlib

from wal.ast_defs import Operator, Symbol, Closure, Environment, Macro, Unquote, UnquoteSplice, WList
from wal.reader import read_wal_sexpr
from wal.passes import expand, optimize, resolve
from wal.util import wal_str


def op_not(seval, args):
    evaluated = seval.eval_args(args)
    assert len(evaluated) >= 1
    assert all(isinstance(arg, int) for arg in evaluated), wal_str(evaluated)
    return not any(evaluated)


def op_eq(seval, args):
    evaluated = seval.eval_args(args)
    assert len(evaluated) > 1
    return all(e == evaluated[0] for e in evaluated)


def op_neq(seval, args):
    evaluated = seval.eval_args(args)
    assert len(evaluated) > 1
    return not all(e == evaluated[0] for e in evaluated)


def op_larger(seval, args):
    assert len(args) == 2
    evaluated = seval.eval_args(args)
    assert all(isinstance(arg, (int, float)) for arg in evaluated)
    return evaluated[0] > evaluated[1]


def op_smaller(seval, args):
    assert len(args) == 2
    evaluated = seval.eval_args(args)
    assert all(isinstance(arg, (int, float)) for arg in evaluated)
    return evaluated[0] < evaluated[1]


def op_larger_equal(seval, args):
    assert len(args) == 2
    evaluated = seval.eval_args(args)
    assert all(isinstance(arg, (int, float)) for arg in evaluated)
    return evaluated[0] >= evaluated[1]


def op_smaller_equal(seval, args):
    assert len(args) == 2
    evaluated = seval.eval_args(args)
    assert all(isinstance(arg, (int, float)) for arg in evaluated)
    return evaluated[0] <= evaluated[1]


def op_and(seval, args):
    assert len(args) > 0, '&&: expects at least one argument'
    for arg in args:
        if not seval.eval(arg):
            return False
    return True


def op_or(seval, args):
    assert len(args) > 0, '||: expects at least one argument'
    for arg in args:
        if seval.eval(arg):
            return True
    return False


def op_let(seval, args):
    assert isinstance(args[0], WList), 'let: expects a list of pairs as first argument'
    save_env = seval.environment
    new_env = Environment(parent=save_env)
    seval.environment = new_env

    for pair in args[0]:
        assert isinstance(pair, WList), 'let: expects a list of pairs as first argument'
        assert isinstance(pair[0], Symbol), 'let: first element of let pair must be a Symbol'
        assert len(pair) == 2, 'let: expects a list of pairs as first argument'
        new_env.define(pair[0].name, seval.eval(pair[1]))

    res = seval.eval_args(args[1:])[-1]
    seval.environment = save_env
    return res


def op_set(seval, args):
    assert args, 'set: expects at least one (key:symbol expr) pair'
    for arg in args:
        assert isinstance(arg, WList), 'set: arguments must be (key:symbol expr) tuples'
        assert len(arg) == 2, 'set: arguments must be (key:symbol expr) tuples'
        key = arg[0]
        assert isinstance(key, Symbol), 'set: key must be a symbol'
        res = seval.eval(arg[1])

        # this signal was already resolved
        if key.steps is not None:
            defined_at = seval.environment
            steps = key.steps
            while steps > 0:
                defined_at = seval.environment.parent
                steps -= 1

            # get the dict out of the environment
            defined_at = defined_at.environment
        else:
            defined_at = seval.environment.is_defined(key.name)

        if defined_at:
            defined_at[key.name] = res
        else:
            assert False, f'Write to undefined symbol {key.name}'

    return res


def op_define(seval, args):
    assert len(args) == 2, 'define: expects exactly two arguments (define key:symbol expr)'
    key = args[0]
    assert isinstance(key, Symbol), 'define: key must be a symbol'
    res = seval.eval(args[1])
    seval.environment.define(key.name, res)
    return res


def op_print(seval, args):
    res = [x if isinstance(x, str) else wal_str(x) for x in seval.eval_args(args)]
    print(*res, sep='')


def op_printf(seval, args):
    assert args, 'printf: expects at least a format string'
    format_evaluated = seval.eval(args[0])
    if isinstance(format_evaluated, str):
        def wal_stringify(expr):
            if isinstance(expr, str):
                return expr
            if isinstance(expr, (list, WList, dict, Operator, Unquote, UnquoteSplice, Macro, Closure, Symbol)):
                return wal_str(expr)

            return expr
        
        evaluated = [wal_stringify(e) for e in seval.eval_args(args[1:])]
        try:
            final_str = format_evaluated % tuple(evaluated)
            print(final_str, sep='', end='')
        except Exception as e:
            raise RuntimeError(str(e) + ': "' + format_evaluated + '"') from e
    else:
        raise ValueError('printf\'s first argument must be a format string')


def op_if(seval, args):
    assert len(args) == 2 or len(args) == 3, 'if: expects a condition, if-clause and optionally an else-clause'
    if seval.eval(args[0]):
        return seval.eval(args[1])

    if len(args) == 3:
        return seval.eval(args[2])

    return None


def op_case(seval, args):
    assert args, 'case: expects at least one case argument'
    keyform = seval.eval(args[0])
    clauses = args[1:]
    if len(set(map(lambda x: str(x[0]), clauses))) != len(clauses):
        raise ValueError('case with duplicate key')

    default = None
    for clause in clauses:
        key = clause[0]
        consequents = clause[1:]

        if keyform == key:
            return seval.eval_args(consequents)[-1]

        if isinstance(key, Symbol) and key.name == 'default':
            default = seval.eval_args(consequents)[-1]

    return default


def op_do(seval, args):
    if args:
        return seval.eval_args(args)[-1]


def op_while(seval, args):
    assert len(args) >= 2, 'while: expects exactly two arguments (while cond (body:expr))'
    condition = args[0]
    body = args[1:]
    res = None
    while seval.eval(condition):
        res = seval.eval_args(body)[-1]
    return res


def op_alias(seval, args):
    assert args, 'alias: expects at least two arguments'
    assert len(args) == 2, 'alias: expects two arguments'

    assert isinstance(args[0], Symbol), 'alias: aliases must be symbols'
    name = seval.eval(args[1])
    assert isinstance(name, (str, Symbol)), 'alias: second argument must evaluate to a string or a symbol'
    if isinstance(name, Symbol):
        name = name.name

    seval.aliases[args[0].name] = name


def op_unalias(seval, args):
    assert len(args) >= 1, 'unalias: expects at least one argument (unalias id:symbol+)'
    for arg in args:
        assert isinstance(arg, Symbol), 'unalias: argument must be a symbol'
        assert arg.name in seval.aliases, f'unalias: no alias {arg.name} known. Can\'t unalias'
        del seval.aliases[arg.name]


def op_quote(seval, args):  # pylint: disable=W0613
    assert len(args) == 1, 'quote: expects exactly one argument'
    return args[0]


def op_quasiquote(seval, args):
    assert len(args) == 1, 'quasiqoute: expects exactly one argument'

    def unquote(expr):
        if isinstance(expr, WList) and len(expr) > 0:
            res = WList([])
            for element in expr:
                if isinstance(element, Unquote):
                    res.append(seval.eval(unquote(element.content)))
                elif isinstance(element, UnquoteSplice):
                    res = res + seval.eval(unquote(element.content))
                else:
                    res.append(unquote(element))

            return res

        return expr

    return unquote(args[0])


def op_unquote(seval, args):
    assert False, 'unquote: not in quasiquote'


def op_eval(seval, args):
    assert len(args) == 1, 'eval: expects exactly one argument'
    evaluated = seval.eval(args[0])
    expanded = expand(seval, evaluated, parent=seval.global_environment)
    optimized = optimize(expanded)
    resolved = resolve(optimized)
    return seval.eval(resolved)


def op_parse(seval, args):
    '''Parses and returns its arguments as WAL expressions'''
    evaluated = seval.eval_args(args)
    assert(all(isinstance(arg, str) for arg in evaluated)), 'parse: all arguments must be string (parse string+)'
    sexprs = [optimize(expand(seval, read_wal_sexpr(arg), parent=seval.global_environment)) for arg in evaluated]
    if len(sexprs) == 1:
        return sexprs[0]

    return WList([Operator.DO] + sexprs)


def op_fn(seval, args):
    assert len(args) >= 2, 'lambda: expects at least two arguments (lambda (symbol | (symbol*)) sexpr)'
    assert isinstance(args[0], (WList, list, Symbol)), 'lambda: first argument must be a list of symbols or a single symbol'
    if isinstance(args[0], (list, WList)):
        assert all(isinstance(arg, Symbol) for arg in args[0]), f'lambda: arguments must be symbols but are {wal_str(args[0])}'

    name = args[1] if isinstance(args[1], (Symbol, str)) else "lambda"

    return Closure(seval.environment, args[0], WList([Operator.DO] + args[1:]), name=name)


def op_defmacro(seval, args):
    assert len(args) >= 3, 'defmacro: expects at least three arguments'
    assert isinstance(args[0], Symbol), f'defmacro: first argument must be a symbol not {args[0]}'
    assert isinstance(args[2], (Symbol, int, str, WList, float)), 'defmacro: third argument must be a valid expression'

    if isinstance(args[0], WList):
        assert all(isinstance(arg, Symbol) for arg in args[0])

    body = WList(list(map(lambda arg: expand(seval, arg), args[2:])))
    seval.environment.define(args[0].name, Macro(args[0].name, args[1], WList([Operator.DO, *body])))


def op_macroexpand(seval, args):
    assert len(args) == 1, 'macroexpand: expects exactly one argument'
    expanded = expand(seval, seval.eval(args[0]), parent=seval.environment)
    optimized = optimize(expanded)
    return optimized


def op_gensym(seval, args):
    seval.gensymi += 1
    return Symbol(f'${seval.gensymi}', 0)


def op_get(seval, args):
    '''get returns the value of a wavefile signal by its string name'''
    assert len(args) == 1, 'get: expects exactly one argument'
    evaluated = seval.eval_args(args)
    if isinstance(evaluated[0], str):
        return seval.eval(Symbol(evaluated[0]))

    if isinstance(evaluated[0], Symbol):
        return seval.eval(evaluated[0])

    return None


def op_import(seval, args):
    '''Evaluate import operations'''
    assert len(args) >= 1, 'Import expects at least one argument'

    def do_import(name):
        sys.path.append(os.getcwd())
        if isinstance(name, str):
            seval.imports[name] = importlib.import_module(name)
        elif isinstance(name, Symbol):
            seval.imports[name.name] = importlib.import_module(name.name)
        else:
            raise ValueError(
                'Import arguments must be a string or a Symbol, not ' + str(type(name)))

    for arg in args:
        if isinstance(arg, Symbol):
            do_import(arg.name)
        else:
            evaluated = seval.eval({}, arg)
            do_import(evaluated)


def op_call(seval, args):
    '''Evaluate calls to external python modules'''
    extern_name = args[0]
    extern_args = args[1:]
    assert isinstance(extern_name, Symbol)
    split = extern_name.name.split('.')
    func = seval.imports[split[0]]

    for fid in split[1:]:
        func = getattr(func, fid)

    evaluated = list(map(seval.eval, extern_args))
    return func(*evaluated)


def op_type(seval, args):
    '''Return the type of argument'''
    return type(seval.eval(args[0]))


def op_rel_eval(seval, args):
    '''Evaluate an expression at a locally modified index. Index is restored after eval is done.'''
    assert len(args) == 2, 'reval: expects two arguments (reval expr:expr offset:expr->int)'
    assert isinstance(args[0], (Symbol, int, str, WList, list, float)), 'reval: first argument must be a valid expression'
    offset = seval.eval(args[1])
    assert isinstance(offset, int), 'reval: second argument must evaluate to int'
    # check if any trace becomes oob with offset
    for trace in seval.traces.traces.values():
        if trace.index + offset > trace.max_index or trace.index + offset < 0:
            return False

    seval.traces.store_indices()
    seval.traces.step(offset)
    res = seval.eval(args[0])
    seval.traces.restore_indices()
    return res


def op_scoped(seval, args):
    assert len(args) == 2, 'scoped: exactly two arguments required (scoped scope:symbol expression)'
    prev_scope = seval.scope

    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (str, Symbol)), 'scoped: argument must be Symbol, string or must evaluate to one of them'

    if isinstance(evaluated, Symbol):
        name = evaluated.name
    elif isinstance(evaluated, str):
        name = evaluated

    seval.scope = name
    seval.global_environment.write('CS', seval.scope)
    res = seval.eval(args[1])
    seval.scope = prev_scope
    seval.global_environment.write('CS', prev_scope)
    return res


def op_allscopes(seval, args):
    assert args, 'all-scopes: exactly one argument required'
    assert isinstance(args[0], WList), 'all-scopes: argument must be a valid expression'
    res = []
    prev_scope = seval.global_environment.read('CS')
    for scope in seval.traces.scopes: # pylint: disable=E1101
        seval.scope = scope
        seval.global_environment.write('CS', scope)
        res.append(seval.eval(args[0]))

    seval.global_environment.write('CS', prev_scope)
    return res


def op_resolve_scope(seval, args):
    assert len(args) == 1, 'resolve-scope: exactly one argument required (resolve-scope name:symbol)'
    assert isinstance(args[0], Symbol), 'resolve-scope: exactly one argument required (resolve-scope name:symbol)'
    if args[0].name in seval.aliases:
        args[0].name = seval.aliases[args[0].name]

    if seval.global_environment.read('CS') in seval.traces.scopes:
        # if the scope is a real scope add .
        name = seval.global_environment.read('CS') + '.' + args[0].name
    else:  # if scope is not a real scope it must be a group name, no dot required
        name = seval.global_environment.read('CS') + args[0].name

    assert seval.traces.contains(name), f'resolve-scope: No signal with name "{name}"'
    return seval.traces.signal_value(name)


def op_set_scope(seval, args):
    assert args, 'set-scope: exactly one argument required'
    assert isinstance(args[0], Symbol), f'set-scope: argument must be Symbol but is {type(args[0]).__name__}'
    assert args[0].name in seval.traces.scopes, f'set-scope: {args[0].name} is not a valid scope' # pylint: disable=E1101
    seval.scope = args[0].name
    seval.global_environment.write('CS', args[0].name)


def op_unset_scope(seval, args):
    assert not args, 'unset-scope: expects no arguments'
    seval.scope = ''
    seval.global_environment.write('CS', '')



def op_groups(seval, args):
    '''
    (group "*_ready" "*_valid" "*_data") => all signals (as scope?) for which p_ready, p_valid. exist
    '''
    assert args, 'groups: expects at least one argument (groups post:str+)'

    def process_arg(arg):
        '''evaluates arg if it is a list, else unpacks a signal or just returns a string'''
        if isinstance(arg, WList):
            return seval.eval(arg)
        if isinstance(arg, Symbol):
            return arg.name
        if isinstance(arg, str):
            return arg

        raise RuntimeError('groups: arguments must be string, symbol, or must evaluate to these types')

    def lookup_alias(arg):
        if arg in seval.aliases:
            return seval.aliases[arg]

        return arg

    args = list(map(lookup_alias, map(process_arg, args)))

    if seval.global_environment.read('CS'):
        pattern = re.compile(rf'{seval.global_environment.read("CS")}\.[^\\.]+{args[0]}')
    else:
        pattern = re.compile(rf'.*{args[0]}')

    candidates = list(filter(pattern.fullmatch, seval.traces.signals))
    groups = set()
    for pre in candidates:
        pre = pre[:-len(args[0])]  # cut off filter suffix
        if all(seval.traces.contains(f'{pre}{post}') for post in args[1:]):
            groups.add(pre)

    groups = list(groups)
    groups.sort()

    return groups


def op_in_group(seval, args):
    assert len(args) >= 2, 'in-group: exactly two arguments required (in-group group:symbol expression)'
    prev_group = seval.group
    prev_scope = seval.scope

    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (str, Symbol)), 'in-group: argument must be Symbol, string or must evaluate to one of them'

    if isinstance(evaluated, Symbol):
        name = evaluated.name
    elif isinstance(evaluated, str):
        name = evaluated

    seval.group = name
    seval.global_environment.write('CG', seval.group)

    scope_index = seval.group.rfind('.')
    seval.scope = seval.group[:scope_index + 1] if scope_index != -1 else prev_scope
    seval.global_environment.write('CS', seval.scope)

    res = seval.eval_args(args[1:])
    seval.group = prev_group
    seval.scope = prev_scope
    seval.global_environment.write('CG', prev_group)
    seval.global_environment.write('CS', prev_scope)
    return res[-1]


def op_in_groups(seval, args):
    assert len(args) >= 2, 'in-groups: exactly two arguments required (in-group group:(symbol+) expression)'
    groups = seval.eval(args[0])
    assert isinstance(groups, (list, WList)), 'in-groups: first argument must evaluate to list'

    res = None
    for group in groups:
        res = op_in_group(seval, [group, *args[1:]])
    return res


def op_resolve_group(seval, args):
    assert len(args) == 1, 'resolve-group (#): exactly one argument required (resolve-group name:symbol)'
    assert isinstance(args[0], Symbol), 'resolve-group: exactly one argument required (resolve-group name:symbol)'
    if args[0].name in seval.aliases:
        args[0].name = seval.aliases[args[0].name]

    name = seval.group + args[0].name

    assert seval.traces.contains(name), f'resolve-group: No signal with name "{name}"'
    return seval.traces.signal_value(name)


def op_slice(seval, args):
    assert len(args) > 1 and len(args) < 4, 'slice: two or three arguments required (slice high:int [low:int])'
    evaluated = seval.eval_args(args)
    assert isinstance(evaluated[0], (int, WList, list, str)), 'slice: first argument must evaluate to a number or a list'

    if isinstance(evaluated[0], int):
        if len(args) == 2: # pylint: disable=R1705
            index = evaluated[1]
            assert isinstance(index, int), 'slice: index must evaluate to int'
            return (evaluated[0] & (1 << index)) >> index
        elif len(args) == 3:
            upper = evaluated[1]
            assert isinstance(upper, int), 'slice: upper index must evaluate to int'
            lower = evaluated[2]
            assert isinstance(lower, int), 'slice: lower index must evaluate to int'
            return (evaluated[0] & (((1 << (upper - lower + 1)) - 1) << lower)) >> lower
    elif isinstance(evaluated[0], (WList, list, str)):
        if len(args) == 2: # pylint: disable=R1705
            index = evaluated[1]
            assert isinstance(index, int), 'slice: index must evaluate to int'
            return evaluated[0][index]
        elif len(args) == 3:
            upper = evaluated[1]
            assert isinstance(upper, int), 'slice: upper index must evaluate to int'
            lower = evaluated[2]
            assert isinstance(lower, int), 'slice: lower index must evaluate to int'
            return evaluated[0][upper:lower]

    return None


def op_loaded_traces(seval, args):
    assert(len(args) == 0), 'loaded-traces: Expects no arguments (loaded-traces)'
    return list(seval.traces.traces.keys())


def op_exit(seval, args):
    assert len(args) < 2, 'exit: expects none or one argument (exit return_code:int)'
    if len(args) == 0:
        sys.exit(0)
    else:
        code = seval.eval(args[0])
        assert isinstance(code, int), 'exit: first argument must evaluate to int'
        sys.exit(code)


core_operators = {
    Operator.NOT.value: op_not,
    Operator.EQ.value: op_eq,
    Operator.NEQ.value: op_neq,
    Operator.LARGER.value: op_larger,
    Operator.SMALLER.value: op_smaller,
    Operator.LARGER_EQUAL.value: op_larger_equal,
    Operator.SMALLER_EQUAL.value: op_smaller_equal,
    Operator.AND.value: op_and,
    Operator.OR.value: op_or,
    Operator.LET.value: op_let,
    Operator.DEFINE.value: op_define,
    Operator.SET.value: op_set,
    Operator.PRINT.value: op_print,
    Operator.PRINTF.value: op_printf,
    Operator.IF.value: op_if,
    Operator.CASE.value: op_case,
    Operator.DO.value: op_do,
    Operator.WHILE.value: op_while,
    Operator.ALIAS.value: op_alias,
    Operator.UNALIAS.value: op_unalias,
    Operator.QUOTE.value: op_quote,
    Operator.QUASIQUOTE.value: op_quasiquote,
    Operator.UNQUOTE.value: op_unquote,
    Operator.EVAL.value: op_eval,
    Operator.PARSE.value: op_parse,
    Operator.DEFMACRO.value: op_defmacro,
    Operator.MACROEXPAND.value: op_macroexpand,
    Operator.GENSYM.value: op_gensym,
    Operator.FN.value: op_fn,
    Operator.GET.value: op_get,
    Operator.IMPORT.value: op_import,
    Operator.CALL.value: op_call,
    Operator.TYPE.value: op_type,
    Operator.REL_EVAL.value: op_rel_eval,
    Operator.SCOPED.value: op_scoped,
    Operator.RESOLVE_SCOPE.value: op_resolve_scope,
    Operator.ALLSCOPES.value: op_allscopes,
    Operator.SETSCOPE.value: op_set_scope,
    Operator.UNSETSCOPE.value: op_unset_scope,
    Operator.GROUPS.value: op_groups,
    Operator.IN_GROUP.value: op_in_group,
    Operator.IN_GROUPS.value: op_in_groups,
    Operator.RESOLVE_GROUP.value: op_resolve_group,
    Operator.SLICE.value: op_slice,
    Operator.LOADED_TRACES.value: op_loaded_traces,
    Operator.EXIT.value: op_exit,
}
